from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import CommandError
from django.db import NotSupportedError
from django.db.backends.base.features import BaseDatabaseFeatures
from django.utils.functional import cached_property
from pymongo.errors import OperationFailure

try:
    from .gis.features import GISFeatures
except ImproperlyConfigured:
    # GIS libraries (GDAL/GEOS) not installed.
    class GISFeatures:
        pass


class DatabaseFeatures(GISFeatures, BaseDatabaseFeatures):
    minimum_database_version = (7, 0)
    allow_sliced_subqueries_with_in = False
    allows_multiple_constraints_on_same_fields = False
    can_create_inline_fk = False
    can_introspect_foreign_keys = False
    can_return_rows_from_bulk_insert = True
    greatest_least_ignores_nulls = True
    has_json_object_function = False
    has_native_json_field = True
    rounds_to_even = True
    supports_boolean_expr_in_select_clause = True
    supports_collation_on_charfield = False
    supports_column_check_constraints = False
    supports_date_lookup_using_string = False
    supports_deferrable_unique_constraints = False
    supports_explaining_query_execution = True
    supports_expression_defaults = False
    supports_expression_indexes = False
    supports_foreign_keys = False
    supports_ignore_conflicts = False
    supports_json_field_contains = False
    # BSON Date type doesn't support microsecond precision.
    supports_microsecond_precision = False
    supports_nulls_distinct_unique_constraints = True
    supports_paramstyle_pyformat = False
    supports_sequence_reset = False
    supports_slicing_ordering_in_compound = True
    supports_table_check_constraints = False
    supports_temporal_subtraction = True
    # MongoDB stores datetimes in UTC.
    supports_timezones = False
    # While MongoDB supports transactions in some configurations (see
    # DatabaseFeatures._supports_transactions), django.db.transaction.atomic() is a
    # no-op on this backend.
    supports_transactions = False
    supports_unspecified_pk = True
    uses_savepoints = False

    _django_test_expected_failures = {
        # $concat only supports strings, not int
        "db_functions.text.test_concat.ConcatTests.test_concat_non_str",
        # QuerySet.order_by() with annotation transform doesn't work:
        # "Expression $mod takes exactly 2 arguments. 1 were passed in"
        # https://github.com/django/django/commit/b0ad41198b3e333f57351e3fce5a1fb47f23f376
        "aggregation.tests.AggregateTestCase.test_order_by_aggregate_transform",
        # 'NulledTransform' object has no attribute 'as_mql'.
        "lookup.tests.LookupTests.test_exact_none_transform",
        # BaseExpression.convert_value() crashes with Decimal128.
        "aggregation.tests.AggregateTestCase.test_combine_different_types",
        "annotations.tests.NonAggregateAnnotationTestCase.test_combined_f_expression_annotation_with_aggregation",
        # Pattern lookups that use regexMatch don't work on JSONField:
        # Unsupported conversion from array to string in $convert
        "model_fields.test_jsonfield.TestQuerying.test_icontains",
        # Unexpected alias_refcount in alias_map.
        "queries.tests.Queries1Tests.test_order_by_tables",
        # Pattern lookups (startswith, regex, etc.) don't work on non-string
        # fields: https://jira.mongodb.org/browse/INTPYTHON-734
        "admin_changelist.tests.ChangeListTests.test_pk_in_search_fields",
        "admin_changelist.tests.ChangeListTests.test_related_field_multiple_search_terms",
        "lookup.tests.LookupTests.test_lookup_int_as_str",
        "lookup.tests.LookupTests.test_regex_non_string",
        # The $sum aggregation returns 0 instead of None for null.
        "aggregation.test_filter_argument.FilteredAggregateTests.test_plain_annotate",
        "aggregation.tests.AggregateTestCase.test_aggregation_default_passed_another_aggregate",
        "aggregation.tests.AggregateTestCase.test_annotation_expressions",
        "aggregation.tests.AggregateTestCase.test_reverse_fkey_annotate",
        "aggregation_regress.tests.AggregationTests.test_annotation_disjunction",
        "aggregation_regress.tests.AggregationTests.test_decimal_aggregate_annotation_filter",
        # subclasses of BaseDatabaseWrapper may require an is_usable() method
        "backends.tests.BackendTestCase.test_is_usable_after_database_disconnects",
        # Connection creation doesn't follow the usual Django API.
        "backends.tests.ThreadTests.test_pass_connection_between_threads",
        "backends.tests.ThreadTests.test_default_connection_thread_local",
        # Object of type ObjectId is not JSON serializable.
        "auth_tests.test_views.LoginTest.test_login_session_without_hash_session_key",
        # GenericRelation.value_to_string() assumes integer pk.
        "contenttypes_tests.test_fields.GenericRelationTests.test_value_to_string",
        # ArrayField's contained_by lookup crashes with Exists: "both operands "
        # of $setIsSubset must be arrays. Second argument is of type: null"
        # https://jira.mongodb.org/browse/SERVER-99186
        "model_fields_.test_arrayfield.QueryingTests.test_contained_by_subquery",
        # Value.as_mql() doesn't call output_field.get_db_prep_save():
        # https://github.com/mongodb/django-mongodb-backend/issues/282
        "model_fields.test_jsonfield.TestSaveLoad.test_bulk_update_custom_get_prep_value",
        # CheckConstraint(condition=models.Q(price__gte=0)) doesn't accept null
        # values because {'$gte': [None, 0]} returns False instead of NULL like
        # in SQL.
        "constraints.tests.CheckConstraintTests.test_validate_nullable_field_with_none",
        # bulk_create() population of _order doesn't work because of ObjectId
        # type mismatch when querying object_id CharField.
        # https://github.com/django/django/commit/953095d1e603fe0f8f01175b1409ca23818dcff9
        "contenttypes_tests.test_order_with_respect_to.OrderWithRespectToGFKTests.test_bulk_create_allows_duplicate_order_values",
        "contenttypes_tests.test_order_with_respect_to.OrderWithRespectToGFKTests.test_bulk_create_mixed_scenario",
        "contenttypes_tests.test_order_with_respect_to.OrderWithRespectToGFKTests.test_bulk_create_respects_mixed_manual_order",
        "contenttypes_tests.test_order_with_respect_to.OrderWithRespectToGFKTests.test_bulk_create_with_existing_children",
    }

    @cached_property
    def django_test_expected_failures(self):
        expected_failures = super().django_test_expected_failures
        expected_failures.update(self._django_test_expected_failures)
        return expected_failures

    _django_test_skips = {
        "Database defaults aren't supported by MongoDB.": {
            # bson.errors.InvalidDocument: cannot encode object:
            # <django.db.models.expressions.DatabaseDefault
            "basic.tests.ModelInstanceCreationTests.test_save_primary_with_db_default",
            "basic.tests.ModelInstanceCreationTests.test_save_primary_with_falsey_db_default",
            "constraints.tests.UniqueConstraintTests.test_database_default",
            "field_defaults.tests.DefaultTests",
            "migrations.test_operations.OperationTests.test_add_field_both_defaults",
            "migrations.test_operations.OperationTests.test_add_field_database_default",
            "migrations.test_operations.OperationTests.test_add_field_database_default_special_char_escaping",
            "migrations.test_operations.OperationTests.test_alter_field_add_database_default",
            "migrations.test_operations.OperationTests.test_alter_field_change_blank_nullable_database_default_to_not_null",
            "migrations.test_operations.OperationTests.test_alter_field_change_default_to_database_default",
            "migrations.test_operations.OperationTests.test_alter_field_change_nullable_to_database_default_not_null",
            "migrations.test_operations.OperationTests.test_alter_field_change_nullable_to_decimal_database_default_not_null",
            "schema.tests.SchemaTests.test_db_default_output_field_resolving",
            "schema.tests.SchemaTests.test_rename_keep_db_default",
            "validation.test_unique.PerformUniqueChecksTest.test_unique_db_default",
        },
        "Insert expressions aren't supported.": {
            "basic.tests.ModelTest.test_save_expressions",
            "bulk_create.tests.BulkCreateTests.test_bulk_insert_now",
            "bulk_create.tests.BulkCreateTests.test_bulk_insert_expressions",
            "expressions.tests.BasicExpressionsTests.test_new_object_create",
            "expressions.tests.BasicExpressionsTests.test_new_object_save",
            "expressions.tests.BasicExpressionsTests.test_object_create_with_aggregate",
            "expressions.tests.BasicExpressionsTests.test_object_create_with_f_expression_in_subquery",
            "expressions.tests.BasicExpressionsTests.test_object_update_unsaved_objects",
            # PI()
            "db_functions.math.test_round.RoundTests.test_decimal_with_precision",
            "db_functions.math.test_round.RoundTests.test_float_with_precision",
        },
        "MongoDB doesn't rename an index when a field is renamed.": {
            "migrations.test_operations.OperationTests.test_rename_field_index_together",
            "migrations.test_operations.OperationTests.test_rename_field_unique_together",
        },
        "AutoField not supported.": {
            "bulk_create.tests.BulkCreateTests.test_bulk_insert_nullable_fields",
            "introspection.tests.IntrospectionTests.test_sequence_list",
            "lookup.tests.LookupTests.test_filter_by_reverse_related_field_transform",
            "lookup.tests.LookupTests.test_in_ignore_none_with_unhashable_items",
            "m2m_through_regress.tests.ThroughLoadDataTestCase.test_sequence_creation",
            "many_to_many.tests.ManyToManyTests.test_add_remove_invalid_type",
            "many_to_one.tests.ManyToOneTests.test_fk_to_smallautofield",
            "many_to_one.tests.ManyToOneTests.test_fk_to_bigautofield",
            "migrations.test_operations.OperationTests.test_autofield__bigautofield_foreignfield_growth",
            "migrations.test_operations.OperationTests.test_model_with_bigautofield",
            "migrations.test_operations.OperationTests.test_smallfield_autofield_foreignfield_growth",
            "migrations.test_operations.OperationTests.test_smallfield_bigautofield_foreignfield_growth",
            "model_fields.test_autofield.AutoFieldTests",
            "model_fields.test_autofield.BigAutoFieldTests",
            "model_fields.test_autofield.SmallAutoFieldTests",
            "queries.tests.TestInvalidValuesRelation.test_invalid_values",
            "schema.tests.SchemaTests.test_alter_autofield_pk_to_bigautofield_pk",
            "schema.tests.SchemaTests.test_alter_autofield_pk_to_smallautofield_pk",
        },
        "Converters aren't run on returning fields from insert.": {
            # Unsure this is needed for this backend. Can implement by request.
            # https://github.com/django/django/commit/d9de74141e8a920940f1b91ed0a3ccb835b55729
            "custom_pk.tests.CustomPKTests.test_auto_field_subclass_bulk_create",
            "custom_pk.tests.CustomPKTests.test_auto_field_subclass_create",
        },
        "MongoDB does not enforce PositiveIntegerField constraint.": {
            "model_fields.test_integerfield.PositiveIntegerFieldTests.test_negative_values",
        },
        "Test assumes integer primary key.": {
            "custom_managers.tests.CustomManagersRegressTestCase.test_save_clears_annotations_from_base_manager",
            "db_functions.comparison.test_cast.CastTests.test_cast_to_integer_foreign_key",
            "expressions.tests.BasicExpressionsTests.test_nested_subquery_outer_ref_with_autofield",
            "model_fields.test_foreignkey.ForeignKeyTests.test_to_python",
            "queries.test_qs_combinators.QuerySetSetOperationTests.test_order_raises_on_non_selected_column",
            "queries.tests.RelatedLookupTypeTests.test_values_queryset_lookup",
            "queries.tests.ValuesSubqueryTests.test_values_in_subquery",
            "sites_tests.tests.CreateDefaultSiteTests.test_no_site_id",
        },
        "Test inspects query for SQL": {
            "aggregation.tests.AggregateAnnotationPruningTests.test_non_aggregate_annotation_pruned",
            "aggregation.tests.AggregateAnnotationPruningTests.test_unreferenced_aggregate_annotation_pruned",
            "aggregation.tests.AggregateAnnotationPruningTests.test_unused_aliased_aggregate_pruned",
            "aggregation.tests.AggregateAnnotationPruningTests.test_referenced_aggregate_annotation_kept",
            "aggregation.tests.AggregateTestCase.test_count_star",
            "delete.tests.DeletionTests.test_only_referenced_fields_selected",
            "expressions.tests.ExistsTests.test_optimizations",
            "lookup.tests.LookupTests.test_in_ignore_none",
            "lookup.tests.LookupTests.test_lookup_direct_value_rhs_unwrapped",
            "lookup.tests.LookupTests.test_textfield_exact_null",
            "many_to_many.tests.ManyToManyQueryTests.test_count_join_optimization_disabled",
            "many_to_many.tests.ManyToManyQueryTests.test_exists_join_optimization_disabled",
            "many_to_many.tests.ManyToManyTests.test_custom_default_manager_exists_count",
            "many_to_one.tests.ManyToOneTests.test_selects",
            "migrations.test_commands.MigrateTests.test_migrate_syncdb_app_label",
            "migrations.test_commands.MigrateTests.test_migrate_syncdb_deferred_sql_executed_with_schemaeditor",
            "model_fields.test_jsonfield.TestQuerying.test_key_sql_injection_escape",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_foreignkey_reverse",
            "prefetch_related.tests.MultiTableInheritanceTest.test_child_link_prefetch",
            "queries.tests.ExistsSql.test_exists",
            "queries.tests.Queries6Tests.test_col_alias_quoted",
            "schema.tests.SchemaTests.test_rename_column_renames_deferred_sql_references",
            "schema.tests.SchemaTests.test_rename_table_renames_deferred_sql_references",
        },
        "Test checks for SQL in str(queryset.query)": {
            "aggregation_regress.tests.AggregationTests.test_more_more5",
            "aggregation_regress.tests.AggregationTests.test_reverse_join_trimming",
            "aggregation_regress.tests.JoinPromotionTests",
            "custom_lookups.tests.YearLteTests",
            "custom_lookups.tests.CustomizedMethodsTests",
            "expressions.tests.BasicExpressionsTests.test_subquery_sql",
            "expressions.tests.BasicExpressionsTests.test_ticket_18375_chained_filters",
            "expressions.tests.BasicExpressionsTests.test_ticket_18375_join_reuse",
            "expressions.tests.BasicExpressionsTests.test_ticket_18375_kwarg_ordering",
            "expressions.tests.BasicExpressionsTests.test_ticket_18375_kwarg_ordering_2",
            "expressions_case.tests.CaseExpressionTests.test_m2m_reuse",
            "generic_relations_regress.tests.GenericRelationTests.test_join_reuse",
            "lookup.tests.LookupTests.test_in_keeps_value_ordering",
            "model_forms.tests.ModelMultipleChoiceFieldTests.test_clean_does_deduplicate_values",
            "model_inheritance.tests.ModelInheritanceTests.test_inherited_ordering_pk_desc",
            "model_inheritance_regress.tests.ModelInheritanceTest.test_inheritance_joins",
            "ordering.tests.OrderingTests.test_order_by_f_expression_duplicates",
            "queries.tests.DisjunctionPromotionTests",
            "queries.tests.JoinReuseTest.test_fk_reuse",
            "queries.tests.JoinReuseTest.test_fk_reuse_annotation",
            "queries.tests.JoinReuseTest.test_fk_reuse_disjunction",
            "queries.tests.JoinReuseTest.test_fk_reuse_order_by",
            "queries.tests.JoinReuseTest.test_fk_reuse_select_related",
            "queries.tests.JoinReuseTest.test_revfk_noreuse",
            "queries.tests.JoinReuseTest.test_revo2o_reuse",
            "queries.tests.NullJoinPromotionOrTest.test_null_join_demotion",
            "queries.tests.NullableRelOrderingTests.test_join_already_in_query",
            "queries.tests.Queries1Tests.test_order_by_join_unref",
            "queries.tests.Queries1Tests.test_subquery_condition",
            "queries.tests.Queries4Tests.test_order_by_resetting",
            "queries.tests.Queries6Tests.test_nested_queries_sql",
            "queries.tests.Queries6Tests.test_ticket_11320",
            "queries.tests.ReverseJoinTrimmingTest.test_reverse_trimming",
            "queries.tests.ValuesJoinPromotionTests.test_non_nullable_fk_not_promoted",
            "queries.tests.ValuesJoinPromotionTests.test_values_no_promotion_for_existing",
            "queries.tests.Ticket18785Tests.test_ticket_18785",
            "select_for_update.tests.SelectForUpdateTests.test_ordered_select_for_update",
            "select_related_regress.tests.SelectRelatedRegressTests.test_null_join_promotion",
            "select_related_regress.tests.SelectRelatedRegressTests.test_regression_7110",
        },
        "Custom aggregations/functions with SQL don't work on MongoDB.": {
            "aggregation.tests.AggregateTestCase.test_add_implementation",
            "aggregation.tests.AggregateTestCase.test_multi_arg_aggregate",
            "aggregation.tests.AggregateTestCase.test_empty_result_optimization",
            "annotations.tests.NonAggregateAnnotationTestCase.test_custom_functions",
            "annotations.tests.NonAggregateAnnotationTestCase.test_custom_functions_can_ref_other_functions",
        },
        "Bilateral transform not implemented.": {
            "db_functions.tests.FunctionTests.test_func_transform_bilateral",
            "db_functions.tests.FunctionTests.test_func_transform_bilateral_multivalue",
        },
        "MongoDB does not support this database function.": {
            "db_functions.math.test_sign.SignTests",
            "db_functions.text.test_chr.ChrTests",
            "db_functions.text.test_md5.MD5Tests",
            "db_functions.text.test_ord.OrdTests",
            "db_functions.text.test_pad.PadTests",
            "db_functions.text.test_repeat.RepeatTests",
            "db_functions.text.test_reverse.ReverseTests",
            "db_functions.text.test_right.RightTests",
            "db_functions.text.test_sha1.SHA1Tests",
            "db_functions.text.test_sha224.SHA224Tests",
            "db_functions.text.test_sha256.SHA256Tests",
            "db_functions.text.test_sha384.SHA384Tests",
            "db_functions.text.test_sha512.SHA512Tests",
        },
        "MongoDB can't annotate ($project) a function like PI().": {
            "aggregation.tests.AggregateTestCase.test_aggregation_default_using_decimal_from_database",
            "db_functions.math.test_pi.PiTests.test",
        },
        "Can't cast from date to datetime without MongoDB interpreting the new value in UTC.": {
            "db_functions.comparison.test_cast.CastTests.test_cast_from_db_date_to_datetime",
            "db_functions.comparison.test_cast.CastTests.test_cast_from_db_datetime_to_time",
        },
        "Casting datetime/timedelta literals has microsecond differences.": {
            "db_functions.comparison.test_cast.CastTests.test_cast_from_python_to_datetime",
            "db_functions.comparison.test_cast.CastTests.test_cast_to_duration",
        },
        "inspectdb is not supported.": {
            "inspectdb.tests.InspectDBTestCase",
            "inspectdb.tests.InspectDBTransactionalTests",
        },
        "DatabaseIntrospection.get_table_description() not supported.": {
            "introspection.tests.IntrospectionTests.test_bigautofield",
            "introspection.tests.IntrospectionTests.test_get_table_description_col_lengths",
            "introspection.tests.IntrospectionTests.test_get_table_description_names",
            "introspection.tests.IntrospectionTests.test_get_table_description_nullable",
            "introspection.tests.IntrospectionTests.test_get_table_description_types",
            "introspection.tests.IntrospectionTests.test_smallautofield",
        },
        "MongoDB can't introspect primary key.": {
            "introspection.tests.IntrospectionTests.test_get_primary_key_column",
            "schema.tests.SchemaTests.test_alter_primary_key_the_same_name",
            "schema.tests.SchemaTests.test_primary_key",
        },
        "Known issue querying JSONField.": {
            # An ExpressionWrapper annotation with KeyTransform followed by
            # .filter(expr__isnull=False) doesn't use KeyTransformIsNull as it
            # needs to.
            "model_fields.test_jsonfield.TestQuerying.test_expression_wrapper_key_transform",
            # There is no way to distinguish between a JSON "null" (represented
            # by Value(None, JSONField())) and a SQL null (queried using the
            # isnull lookup). Both of these queries return both nulls.
            "model_fields.test_jsonfield.TestSaveLoad.test_json_null_different_from_sql_null",
            # Some queries with Q objects, e.g. Q(value__foo="bar"), don't work
            # properly, particularly with QuerySet.exclude().
            "model_fields.test_jsonfield.TestQuerying.test_lookup_exclude",
            "model_fields.test_jsonfield.TestQuerying.test_lookup_exclude_nonexistent_key",
            # Queries like like QuerySet.filter(value__j=None) incorrectly
            # returns objects where the key doesn't exist.
            "model_fields.test_jsonfield.TestQuerying.test_none_key",
            "model_fields.test_jsonfield.TestQuerying.test_none_key_exclude",
        },
        "Queries without a collection aren't supported on MongoDB.": {
            "queries.test_q.QCheckTests",
            "queries.test_query.TestQueryNoModel",
        },
        "MongoDB doesn't use CursorDebugWrapper.": {
            # It prevents RuntimeWarning for import-time queries from working.
            # https://github.com/django/django/commit/fbd16438f46bc2128926958ad24331da5d1b406f
            "apps.tests.QueryPerformingAppTests",
            "backends.tests.LastExecutedQueryTest.test_last_executed_query",
            "backends.tests.LastExecutedQueryTest.test_last_executed_query_with_duplicate_params",
            "backends.tests.LastExecutedQueryTest.test_query_encoding",
        },
        "Test not applicable for MongoDB's SQLCompiler.": {
            "queries.test_iterator.QuerySetIteratorTests",
        },
        "Support for views not implemented.": {
            "introspection.tests.IntrospectionTests.test_table_names_with_views",
        },
        "Connection health checks not implemented.": {
            "backends.base.test_base.ConnectionHealthChecksTests",
        },
        "transaction.atomic() is not supported.": {
            "backends.base.test_base.DatabaseWrapperLoggingTests",
            "basic.tests.SelectOnSaveTests.test_select_on_save_lying_update",
            "migrations.test_executor.ExecutorTests.test_atomic_operation_in_non_atomic_migration",
            "migrations.test_operations.OperationTests.test_run_python_atomic",
        },
        "transaction.rollback() is not supported.": {
            "transactions.tests.AtomicMiscTests.test_mark_for_rollback_on_error_in_autocommit",
            "transactions.tests.AtomicMiscTests.test_mark_for_rollback_on_error_in_transaction",
            "transactions.tests.NonAutocommitTests.test_orm_query_after_error_and_rollback",
        },
        "migrate --fake-initial is not supported.": {
            "migrations.test_commands.MigrateTests.test_migrate_fake_initial",
            "migrations.test_commands.MigrateTests.test_migrate_fake_split_initial",
            "migrations.test_executor.ExecutorTests.test_soft_apply",
        },
        "Test works in isolation but fails on CI.": {
            # Probably something to do with lack of transaction support.
            "migration_test_data_persistence.tests.MigrationDataNormalPersistenceTestCase.test_persistence",
        },
        "Database caching not implemented.": {
            "cache.tests.CreateCacheTableForDBCacheTests",
            "cache.tests.DBCacheTests",
            "cache.tests.DBCacheWithTimeZoneTests",
        },
        "FilteredRelation not supported.": {
            # https://github.com/mongodb/django-mongodb-backend/issues/157
            "filtered_relation.tests.FilteredRelationAggregationTests",
            "filtered_relation.tests.FilteredRelationAnalyticalAggregationTests",
            "filtered_relation.tests.FilteredRelationTests",
            "queryset_pickle.tests.PickleabilityTestCase.test_pickle_filteredrelation",
        },
        "Broken test.": {
            # This test uses a database router that only allows the auth_user
            # table to be created on "other", but then writes users to
            # "default". It can raise IntegrityError when run alongside other
            # tests because the auth_user table isn't flushed between tests.
            "multiple_database.tests.AuthTestCase.test_dumpdata",
        },
        "ForeignObject is not supported.": {
            "foreign_object.test_agnostic_order_trimjoin.TestLookupQuery.test_deep_mixed_backward",
            "foreign_object.test_empty_join.RestrictedConditionsTests",
            "foreign_object.tests.MultiColumnFKTests",
            "foreign_object.tests.TestExtraJoinFilterQ",
        },
        "Tuple lookups are not supported.": {
            "foreign_object.test_tuple_lookups.TupleLookupsTests",
        },
        "ColPairs is not supported.": {
            "auth_tests.test_views.CustomUserCompositePrimaryKeyPasswordResetTest",
            "composite_pk.test_aggregate.CompositePKAggregateTests",
            "composite_pk.test_create.CompositePKCreateTests",
            "composite_pk.test_delete.CompositePKDeleteTests",
            "composite_pk.test_filter.CompositePKFilterTests",
            "composite_pk.test_get.CompositePKGetTests",
            "composite_pk.test_models.CompositePKModelsTests",
            "composite_pk.test_order_by.CompositePKOrderByTests",
            "composite_pk.test_update.CompositePKUpdateTests",
            "composite_pk.test_values.CompositePKValuesTests",
            "composite_pk.tests.CompositePKTests.test_get_primary_key_columns",
            "composite_pk.tests.CompositePKFixturesTests",
        },
        "Custom lookups are not supported.": {
            "custom_lookups.tests.BilateralTransformTests",
            "custom_lookups.tests.LookupTests.test_basic_lookup",
            "custom_lookups.tests.LookupTests.test_custom_lookup_with_subquery",
            "custom_lookups.tests.LookupTests.test_custom_name_lookup",
            "custom_lookups.tests.LookupTests.test_div3_extract",
            "custom_lookups.tests.SubqueryTransformTests.test_subquery_usage",
        },
        "connection.close() does not close the connection.": {
            "servers.test_liveserverthread.LiveServerThreadTest.test_closes_connections",
            "servers.tests.LiveServerTestCloseConnectionTest.test_closes_connections",
        },
        "Disallowed query protection doesn't work on MongoDB.": {
            # Because this backend doesn't use cursor(), chunked_cursor(), etc.
            # https://github.com/django/django/blob/045110ff3089aefd9c3e65c707df465bacfed986/django/test/testcases.py#L195-L206
            "test_utils.test_testcase.TestTestCase.test_disallowed_database_queries",
            "test_utils.test_transactiontestcase.DisallowedDatabaseQueriesTests.test_disallowed_database_queries",
            "test_utils.tests.DisallowedDatabaseQueriesTests.test_disallowed_database_chunked_cursor_queries",
            "test_utils.tests.DisallowedDatabaseQueriesTests.test_disallowed_database_queries",
            "test_utils.tests.DisallowedDatabaseQueriesTests.test_disallowed_thread_database_connection",
        },
        "search lookup not supported on non-Atlas until MongoDB 8.2.": {
            "expressions.tests.BasicExpressionsTests.test_lookups_subquery",
        },
        "MongoDB does not support cursor.execute().": {
            # Test catches another exception with assertRaises.
            "backends.tests.BackendTestCase.test_duplicate_table_error",
            # AssertionError: <django_mongodb_backend.base.Cursor object at 0x77a8c8b24350>
            # is not an instance of <class 'django.db.backends.utils.CursorWrapper'>
            "backends.tests.BackendTestCase.test_cursor_contextmanager",
            # Many test methods that don't all need to be added to
            # django_test_expected_raises.
            "backends.base.test_base.ExecuteWrapperTests",
            # Test doesn't cleanup properly when it fails, causing a cascade of
            # failures in other tests.
            "migrations.test_commands.MigrateTests.test_migrate_plan",
        },
        "MongoDB does not support cursor.callproc().": {
            # Test catches another exception with assertRaises.
            "backends.test_utils.CursorWrapperTests.test_unsupported_callproc_kparams_raises_error",
        },
        "RawSQL is not supported on MongoDB.": {
            # django_test_expected_raises doesn't work with subTest.
            "annotations.tests.NonAggregateAnnotationTestCase.test_raw_sql_with_inherited_field",
            "expressions.tests.BasicExpressionsTests.test_order_by_multiline_sql",
            # django_test_expected_raises doesn't work with async def.
            "async.test_async_queryset.AsyncQuerySetTest.test_raw",
        },
        "QuerySet.prefetch_related() is not supported on MongoDB.": {
            # django_test_expected_raises doesn't work with subTest.
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_m2m_forward",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_prefetch_related_objects_with_various_iterables",
            "prefetch_related.tests.CustomPrefetchTests.test_filter_deferred",
            "prefetch_related.tests.PrefetchRelatedTests.test_filter_deferred",
        },
    }

    @cached_property
    def django_test_skips(self):
        skips = super().django_test_skips
        skips.update(self._django_test_skips)
        return skips

    # Tests that are expected to raise certain exceptions:
    _django_test_expected_raises = {
        (NotSupportedError, "Pattern lookups on UUIDField are not supported."): {
            "model_fields.test_uuid.TestQuerying.test_contains",
            "model_fields.test_uuid.TestQuerying.test_endswith",
            "model_fields.test_uuid.TestQuerying.test_filter_with_expr",
            "model_fields.test_uuid.TestQuerying.test_icontains",
            "model_fields.test_uuid.TestQuerying.test_iendswith",
            "model_fields.test_uuid.TestQuerying.test_iexact",
            "model_fields.test_uuid.TestQuerying.test_istartswith",
            "model_fields.test_uuid.TestQuerying.test_startswith",
        },
        (
            CommandError,
            "Unable to serialize database: QuerySet.prefetch_related() is not "
            "supported on MongoDB.",
        ): {
            "fixtures.tests.FixtureLoadingTests.test_dumpdata_objects_with_prefetch_related",
        },
        (NotSupportedError, "QuerySet.prefetch_related() is not supported on MongoDB."): {
            "backends.base.test_creation.TestDeserializeDbFromString.test_serialize_db_to_string_base_manager_with_prefetch_related",
            "fixtures.tests.FixtureLoadingTests.test_loading_and_dumping",
            "m2m_through_regress.test_multitable.MultiTableTests.test_m2m_prefetch_proxied",
            "m2m_through_regress.test_multitable.MultiTableTests.test_m2m_prefetch_reverse_proxied",
            "many_to_many.tests.ManyToManyQueryTests.test_prefetch_related_no_queries_optimization_disabled",
            "many_to_many.tests.ManyToManyTests.test_add_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_add_then_remove_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_clear_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_create_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_remove_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_set_after_prefetch",
            "model_forms.tests.OtherModelFormTests.test_prefetch_related_queryset",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_m2m_reverse",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_m2m_then_m2m",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_prefetch_object",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_prefetch_object_to_attr",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_prefetch_object_to_attr_twice",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_prefetch_object_twice",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests.test_prefetch_queryset",
            "prefetch_related.test_uuid.UUIDPrefetchRelated.test_prefetch_related_from_uuid_model",
            "prefetch_related.test_uuid.UUIDPrefetchRelated.test_prefetch_related_from_uuid_model_to_uuid_model",
            "prefetch_related.test_uuid.UUIDPrefetchRelated.test_prefetch_related_to_uuid_model",
            "prefetch_related.test_uuid.UUIDPrefetchRelatedLookups.test_from_integer_pk_lookup_integer_pk_uuid_pk_uuid_pk",
            "prefetch_related.test_uuid.UUIDPrefetchRelatedLookups.test_from_integer_pk_lookup_uuid_pk_integer_pk",
            "prefetch_related.test_uuid.UUIDPrefetchRelatedLookups.test_from_uuid_pk_lookup_integer_pk2_uuid_pk2",
            "prefetch_related.test_uuid.UUIDPrefetchRelatedLookups.test_from_uuid_pk_lookup_uuid_pk_integer_pk",
            "prefetch_related.tests.CustomPrefetchTests.test_ambiguous",
            "prefetch_related.tests.CustomPrefetchTests.test_custom_qs",
            "prefetch_related.tests.CustomPrefetchTests.test_m2m",
            "prefetch_related.tests.CustomPrefetchTests.test_m2m_through_fk",
            "prefetch_related.tests.CustomPrefetchTests.test_nested_prefetch_related_are_not_overwritten",
            "prefetch_related.tests.CustomPrefetchTests.test_nested_prefetch_related_with_duplicate_prefetch_and_depth",
            "prefetch_related.tests.CustomPrefetchTests.test_nested_prefetch_related_with_duplicate_prefetcher",
            "prefetch_related.tests.CustomPrefetchTests.test_o2m_through_m2m",
            "prefetch_related.tests.CustomPrefetchTests.test_reverse_m2m",
            "prefetch_related.tests.CustomPrefetchTests.test_to_attr_cached_property",
            "prefetch_related.tests.CustomPrefetchTests.test_traverse_multiple_items_property",
            "prefetch_related.tests.CustomPrefetchTests.test_traverse_qs",
            "prefetch_related.tests.CustomPrefetchTests.test_traverse_single_item_property",
            "prefetch_related.tests.DefaultManagerTests.test_m2m_then_m2m",
            "prefetch_related.tests.ForeignKeyToFieldTest.test_m2m",
            "prefetch_related.tests.GenericRelationTests.test_traverse_GFK",
            "prefetch_related.tests.LookupOrderingTest.test_order",
            "prefetch_related.tests.MultiDbTests.test_using_is_honored_m2m",
            "prefetch_related.tests.MultiTableInheritanceTest.test_m2m_to_inheriting_model",
            "prefetch_related.tests.PrefetchRelatedMTICacheTests.test_add_clears_prefetched_objects_in_grandparent",
            "prefetch_related.tests.PrefetchRelatedMTICacheTests.test_add_clears_prefetched_objects_in_parent",
            "prefetch_related.tests.PrefetchRelatedMTICacheTests.test_grandparent_m2m_available_in_child",
            "prefetch_related.tests.PrefetchRelatedMTICacheTests.test_parent_m2m_available_in_child",
            "prefetch_related.tests.PrefetchRelatedMTICacheTests.test_remove_clears_prefetched_objects_in_grandparent",
            "prefetch_related.tests.PrefetchRelatedMTICacheTests.test_remove_clears_prefetched_objects_in_parent",
            "prefetch_related.tests.PrefetchRelatedTests.test_attribute_error",
            "prefetch_related.tests.PrefetchRelatedTests.test_foreign_key_then_m2m",
            "prefetch_related.tests.PrefetchRelatedTests.test_forward_m2m_to_attr_conflict",
            "prefetch_related.tests.PrefetchRelatedTests.test_get",
            "prefetch_related.tests.PrefetchRelatedTests.test_invalid_final_lookup",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_forward",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_join_reuse",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_prefetching_iterator_with_chunks",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_reverse",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_then_m2m",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_then_m2m_object_ids",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_then_reverse_fk_object_ids",
            "prefetch_related.tests.PrefetchRelatedTests.test_m2m_then_reverse_one_to_one_object_ids",
            "prefetch_related.tests.PrefetchRelatedTests.test_overriding_prefetch",
            "prefetch_related.tests.PrefetchRelatedTests.test_reverse_m2m_to_attr_conflict",
            "prefetch_related.tests.PrefetchRelatedTests.test_reverse_one_to_one_then_m2m",
            "prefetch_related.tests.ReadPrefetchedObjectsCacheTests.test_retrieves_results_from_prefetched_objects_cache",
            "prefetch_related.tests.Ticket21410Tests.test_bug",
            "queryset_pickle.tests.PickleabilityTestCase.test_pickle_prefetch_related_with_m2m_and_objects_deletion",
            "serializers.test_json.JsonSerializerTestCase.test_serialize_prefetch_related_m2m",
            "serializers.test_json.JsonSerializerTestCase.test_serialize_prefetch_related_m2m_with_natural_keys",
            "serializers.test_jsonl.JsonlSerializerTestCase.test_serialize_prefetch_related_m2m",
            "serializers.test_jsonl.JsonlSerializerTestCase.test_serialize_prefetch_related_m2m_with_natural_keys",
            "serializers.test_xml.XmlSerializerTestCase.test_serialize_prefetch_related_m2m",
            "serializers.test_xml.XmlSerializerTestCase.test_serialize_prefetch_related_m2m_with_natural_keys",
            "serializers.test_yaml.YamlSerializerTestCase.test_serialize_prefetch_related_m2m",
            "serializers.test_yaml.YamlSerializerTestCase.test_serialize_prefetch_related_m2m_with_natural_keys",
        },
        (
            NotSupportedError,
            "Cannot use QuerySet.update() when querying across multiple collections on MongoDB.",
        ): {
            "custom_managers.tests.CustomManagersRegressTestCase.test_refresh_from_db_when_default_manager_filters",
            "expressions.tests.BasicExpressionsTests.test_filter_with_join",
            "model_inheritance.tests.ModelInheritanceDataTests.test_update_works_on_parent_and_child_models_at_once",
            "queries.tests.Queries4Tests.test_ticket7095",
            "queries.tests.Queries5Tests.test_ticket9848",
            "update.tests.AdvancedTests.test_update_annotated_multi_table_queryset",
            "update.tests.AdvancedTests.test_update_ordered_by_m2m_annotation",
            "update.tests.AdvancedTests.test_update_ordered_by_m2m_annotation_desc",
            "update.tests.AdvancedTests.test_update_values_annotation",
        },
        (
            NotSupportedError,
            "Cannot use QuerySet.delete() when querying across multiple collections on MongoDB.",
        ): {
            "admin_changelist.tests.ChangeListTests.test_distinct_for_many_to_many_at_second_level_in_search_fields",
            "admin_changelist.tests.ChangeListTests.test_distinct_for_through_m2m_at_second_level_in_list_filter",
            "delete.tests.FastDeleteTests.test_fast_delete_aggregation",
            "delete.tests.FastDeleteTests.test_fast_delete_empty_no_update_can_self_select",
            "delete.tests.FastDeleteTests.test_fast_delete_full_match",
            "delete.tests.FastDeleteTests.test_fast_delete_joined_qs",
            "delete_regress.tests.DeleteTests.test_meta_ordered_delete",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_annotate",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_defer",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_select_related",
            "one_to_one.tests.OneToOneTests.test_o2o_primary_key_delete",
        },
        (NotSupportedError, "Cannot use QuerySet.delete() when a subquery is required."): {
            "custom_managers.tests.CustomManagerTests.test_removal_through_default_m2m_related_manager",
            "custom_managers.tests.CustomManagerTests.test_removal_through_specified_m2m_related_manager",
            "delete_regress.tests.DeleteTests.test_self_reference_with_through_m2m_at_second_level",
            "many_to_many.tests.ManyToManyTests.test_assign",
            "many_to_many.tests.ManyToManyTests.test_assign_ids",
            "many_to_many.tests.ManyToManyTests.test_clear",
            "many_to_many.tests.ManyToManyTests.test_remove",
            "many_to_many.tests.ManyToManyTests.test_reverse_assign_with_queryset",
            "many_to_many.tests.ManyToManyTests.test_set",
            "many_to_many.tests.ManyToManyTests.test_set_existing_different_type",
        },
        (NotSupportedError, "QuerySet.extra() is not supported on MongoDB."): {
            "aggregation.tests.AggregateTestCase.test_exists_extra_where_with_aggregate",
            "annotations.tests.NonAggregateAnnotationTestCase.test_column_field_ordering",
            "annotations.tests.NonAggregateAnnotationTestCase.test_column_field_ordering_with_deferred",
            "basic.tests.ModelTest.test_extra_method_select_argument_with_dashes",
            "basic.tests.ModelTest.test_extra_method_select_argument_with_dashes_and_values",
            "defer.tests.DeferTests.test_defer_extra",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_extra",
            "extra_regress.tests.ExtraRegressTests",
            "lookup.tests.LookupTests.test_values",
            "lookup.tests.LookupTests.test_values_list",
            "ordering.tests.OrderingTests.test_extra_ordering",
            "ordering.tests.OrderingTests.test_extra_ordering_quoting",
            "queries.test_qs_combinators.QuerySetSetOperationTests.test_union_multiple_models_with_values_list_and_order_by_extra_select",
            "queries.test_qs_combinators.QuerySetSetOperationTests.test_union_with_extra_and_values_list",
            "queries.tests.EscapingTests.test_ticket_7302",
            "queries.tests.Queries1Tests.test_tickets_1878_2939",
            "queries.tests.Queries1Tests.test_tickets_7087_12242",
            "queries.tests.Queries5Tests.test_extra_select_literal_percent_s",
            "queries.tests.Queries5Tests.test_ticket7256",
            "queries.tests.ValuesQuerysetTests.test_extra_multiple_select_params_values_order_by",
            "queries.tests.ValuesQuerysetTests.test_extra_select_params_values_order_in_extra",
            "queries.tests.ValuesQuerysetTests.test_extra_values",
            "queries.tests.ValuesQuerysetTests.test_extra_values_list",
            "queries.tests.ValuesQuerysetTests.test_extra_values_order_multiple",
            "queries.tests.ValuesQuerysetTests.test_extra_values_order_twice",
            "queries.tests.ValuesQuerysetTests.test_flat_extra_values_list",
            "select_related.tests.SelectRelatedTests.test_select_related_with_extra",
        },
        (NotSupportedError, "RawSQL is not supported on MongoDB."): {
            "aggregation.tests.AggregateTestCase.test_coalesced_empty_result_set",
            "aggregation_regress.tests.AggregationTests.test_annotate_with_extra",
            "aggregation_regress.tests.AggregationTests.test_annotation",
            "aggregation_regress.tests.AggregationTests.test_more_more3",
            "aggregation_regress.tests.AggregationTests.test_more_more_more3",
            "async.test_async_queryset.AsyncQuerySetTest.test_raw",
            "custom_methods.tests.MethodsTests.test_custom_methods",
            "expressions.tests.BasicExpressionsTests.test_annotate_values_filter",
            "expressions.tests.BasicExpressionsTests.test_filtering_on_rawsql_that_is_boolean",
            "model_fields.test_jsonfield.TestQuerying.test_key_transform_raw_expression",
            "model_fields.test_jsonfield.TestQuerying.test_nested_key_transform_raw_expression",
            "ordering.tests.OrderingTests.test_extra_ordering_with_table_name",
            "queries.tests.Queries1Tests.test_order_by_rawsql",
            "queries.tests.ValuesQuerysetTests.test_named_values_list_with_fields",
            "queries.tests.ValuesQuerysetTests.test_named_values_list_without_fields",
            "raw_query.tests.RawQueryTests",
        },
        (NotImplementedError, ""): {
            # SchemaEditor.quote_value() isn't implemented.
            "schema.test_logging.SchemaLoggerTests.test_extra_args"
        },
        (NotSupportedError, "MongoDB does not support cursor.execute()."): {
            "backends.tests.BackendTestCase.test_queries",
            "backends.tests.BackendTestCase.test_queries_bare_where",
            "backends.tests.BackendTestCase.test_queries_limit",
            "backends.tests.BackendTestCase.test_queries_logger",
            "backends.tests.BackendTestCase.test_unicode_fetches",
            "backends.tests.EscapingChecks.test_parameter_escaping",
            "backends.tests.EscapingChecks.test_paramless_no_escaping",
            "composite_pk.tests.CompositePKTests.test_raw",
            "composite_pk.tests.CompositePKTests.test_raw_missing_PK_fields",
            "delete_regress.tests.DeleteLockingTest.test_concurrent_delete",
            "migrations.test_multidb.MultiDBOperationTests.test_run_sql_migrate_foo_router_with_hints",
            "migrations.test_operations.OperationTests.test_run_sql",
            "migrations.test_operations.OperationTests.test_run_sql_params",
            "migrations.test_operations.OperationTests.test_separate_database_and_state",
            "multiple_database.tests.QueryTestCase.test_raw",
            "prefetch_related.tests.RawQuerySetTests.test_prefetch_before_raw",
            "prefetch_related.tests.RawQuerySetTests.test_basic",
            "prefetch_related.tests.RawQuerySetTests.test_clear",
            "schema.tests.SchemaTests.test_remove_constraints_capital_letters",
            "test_utils.tests.AllowedDatabaseQueriesTests.test_allowed_database_copy_queries",
            "timezones.tests.LegacyDatabaseTests.test_cursor_execute_accepts_naive_datetime",
            "timezones.tests.LegacyDatabaseTests.test_cursor_execute_returns_naive_datetime",
            "timezones.tests.LegacyDatabaseTests.test_raw_sql",
            "timezones.tests.NewDatabaseTests.test_cursor_execute_accepts_naive_datetime",
            "timezones.tests.NewDatabaseTests.test_cursor_execute_returns_naive_datetime",
            "timezones.tests.NewDatabaseTests.test_cursor_explicit_time_zone",
            "timezones.tests.NewDatabaseTests.test_raw_sql",
        },
        (NotSupportedError, "MongoDB does not support cursor.executemany()."): {
            "backends.tests.BackendTestCase.test_cursor_executemany",
            "backends.tests.BackendTestCase.test_cursor_executemany_with_empty_params_list",
            "backends.tests.BackendTestCase.test_cursor_executemany_with_iterator",
        },
        (
            NotSupportedError,
            "TruncDate with tzinfo (America/Vancouver) isn't supported on MongoDB.",
        ): {
            "model_fields.test_datetimefield.DateTimeFieldTests.test_lookup_date_with_use_tz",
        },
        (NotSupportedError, "TruncDate with tzinfo (Africa/Nairobi) isn't supported on MongoDB."): {
            "timezones.tests.NewDatabaseTests.test_query_convert_timezones",
        },
        (NotSupportedError, "ExtractQuarter is not supported."): {
            "db_functions.datetime.test_extract_trunc.DateFunctionTests.test_extract_quarter_func",
            "db_functions.datetime.test_extract_trunc.DateFunctionTests.test_extract_quarter_func_boundaries",
        },
        (NotSupportedError, "StringAgg is not supported."): {
            "aggregation.tests.AggregateTestCase.test_distinct_on_stringagg",
            "aggregation.tests.AggregateTestCase.test_string_agg_escapes_delimiter",
            "aggregation.tests.AggregateTestCase.test_string_agg_filter",
            "aggregation.tests.AggregateTestCase.test_string_agg_filter_in_subquery",
            "aggregation.tests.AggregateTestCase.test_stringagg_default_value",
        },
        (NotSupportedError, "ColPairs is not supported."): {
            "composite_pk.tests.CompositePKTests.test_in_bulk",
            "composite_pk.tests.CompositePKTests.test_in_bulk_batching",
            "composite_pk.tests.CompositePKTests.test_only",
            "composite_pk.tests.CompositePKTests.test_pk_must_be_list_or_tuple",
            "composite_pk.tests.CompositePKTests.test_pk_updated_if_field_updated",
            "composite_pk.tests.CompositePKTests.test_query",
            "composite_pk.tests.CompositePKTests.test_select_related",
        },
        (NotSupportedError, "MongoDB does not support the 'startswith' lookup in indexes."): {
            "schema.tests.SchemaTests.test_unique_constraint_nulls_distinct_condition",
        },
    }

    @cached_property
    def django_test_expected_raises(self):
        try:
            expected_raises = super().django_test_expected_raises
        except AttributeError:
            # No extra tests if GIS backend is inactive.
            return self._django_test_expected_raises
        # Combine GIS backend's tests with this class's.
        for (exception, msg), tests in self._django_test_expected_raises.items():
            try:
                expected_raises[exception, msg].update(tests)
            except KeyError:
                expected_raises[exception, msg] = tests
        return expected_raises

    @cached_property
    def mongodb_version(self):
        return self.connection.get_database_version()  # e.g., (6, 3, 0)

    @cached_property
    def is_mongodb_8_0(self):
        return self.mongodb_version >= (8, 0)

    @cached_property
    def supports_search(self):
        """Does the server support MongoDB search queries and indexes?"""
        try:
            self.connection.get_collection("__null").list_search_indexes()
        except OperationFailure:
            # It would be best to check the error message or error code to
            # avoid hiding some other exception, but the message/code varies
            # across MongoDB versions. Example error message:
            # "$listSearchIndexes  stage is only allowed on MongoDB Atlas".
            return False
        else:
            return True

    @cached_property
    def _supports_transactions(self):
        """
        Transactions are enabled if MongoDB is configured as a replica set or a
        sharded cluster. This is a MongoDB-specific feature flag distinct from Django's
        supports_transactions (without the underscore prefix).
        """
        self.connection.ensure_connection()
        client = self.connection.connection.admin
        hello = client.command("hello")
        # a replica set or a sharded cluster
        return "setName" in hello or hello.get("msg") == "isdbgrid"

    @cached_property
    def supports_queryable_encryption(self):
        """
        For testing purposes, Queryable Encryption requires a MongoDB 8.0 or
        later replica set or sharded cluster, as well as MongoDB Atlas or
        Enterprise. This flag must not guard any non-test functionality since
        it would prevent MongoDB 7.0 from being used, which also supports
        Queryable Encryption. The models in tests/encryption_ aren't compatible
        with MongoDB 7.0 because {"queryType": "range"} is "rangePreview".
        """
        self.connection.ensure_connection()
        build_info = self.connection.connection.admin.command("buildInfo")
        is_enterprise = "enterprise" in build_info.get("modules")
        return (
            # self.supports_search is a proxy for "is Atlas" until MongoDB 8.2
            # when search support was added in Community Edition. Thus, this
            # may need to be reworked, however, it only affects when tests run.
            (is_enterprise or self.supports_search)
            and self._supports_transactions
            and self.is_mongodb_8_0
        )



from django.core.checks import Error
from django.db import models
from django.test import SimpleTestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.fields import (
    EmbeddedModelArrayField,
    EncryptedEmbeddedModelArrayField,
    EncryptedEmbeddedModelField,
    EncryptedIntegerField,
)
from django_mongodb_backend.models import EmbeddedModel


@isolate_apps("encryption_")
class InvalidModelsTests(SimpleTestCase):
    def test_encrypted_field_in_embedded_model_array_field(self):
        class Embedded(EmbeddedModel):
            subfield = EncryptedIntegerField()

        class Model(models.Model):
            field = EmbeddedModelArrayField(Embedded)

        self.assertEqual(
            Model.check(),
            [
                Error(
                    "EmbeddedModelArrayField cannot contain encrypted "
                    "fields (found EncryptedIntegerField).",
                    obj=Model._meta.get_field("field"),
                    id="mongodb.fields.encryption.E001",
                ),
            ],
        )

    def test_encrypted_field_in_encrypted_embedded_model_field(self):
        class Embedded(EmbeddedModel):
            subfield = EncryptedIntegerField()

        class Model(models.Model):
            field = EncryptedEmbeddedModelField(Embedded)

        self.assertEqual(
            Model.check(),
            [
                Error(
                    "EncryptedEmbeddedModelField cannot contain encrypted "
                    "fields (found EncryptedIntegerField).",
                    obj=Model._meta.get_field("field"),
                    id="mongodb.fields.encryption.E001",
                ),
            ],
        )

    def test_encrypted_field_in_encrypted_embedded_model_array_field(self):
        class Embedded(EmbeddedModel):
            subfield = EncryptedIntegerField()

        class Model(models.Model):
            field = EncryptedEmbeddedModelArrayField(Embedded)

        self.assertEqual(
            Model.check(),
            [
                Error(
                    "EncryptedEmbeddedModelArrayField cannot contain "
                    "encrypted fields (found EncryptedIntegerField).",
                    obj=Model._meta.get_field("field"),
                    id="mongodb.fields.encryption.E001",
                ),
            ],
        )



#!/usr/bin/env python
import os
import sys

import django
from django.core.management import call_command

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "perf.settings")


if __name__ == "__main__":
    django.setup()
    call_command("test", sys.argv[1:])



from django.db import DatabaseError, connection
from django.test import TransactionTestCase, skipIfDBFeature, skipUnlessDBFeature

from django_mongodb_backend import transaction

from .models import Reporter


@skipUnlessDBFeature("_supports_transactions")
class AtomicTests(TransactionTestCase):
    """Largely copied from Django's test/transactions."""

    available_apps = ["transactions_"]

    def test_decorator_syntax_commit(self):
        @transaction.atomic
        def make_reporter():
            return Reporter.objects.create(first_name="Tintin")

        reporter = make_reporter()
        self.assertSequenceEqual(Reporter.objects.all(), [reporter])

    def test_decorator_syntax_rollback(self):
        @transaction.atomic
        def make_reporter():
            Reporter.objects.create(first_name="Haddock")
            raise Exception("Oops, that's his last name")

        with self.assertRaisesMessage(Exception, "Oops"):
            make_reporter()
        self.assertSequenceEqual(Reporter.objects.all(), [])

    def test_alternate_decorator_syntax_commit(self):
        @transaction.atomic()
        def make_reporter():
            return Reporter.objects.create(first_name="Tintin")

        reporter = make_reporter()
        self.assertSequenceEqual(Reporter.objects.all(), [reporter])

    def test_alternate_decorator_syntax_rollback(self):
        @transaction.atomic()
        def make_reporter():
            Reporter.objects.create(first_name="Haddock")
            raise Exception("Oops, that's his last name")

        with self.assertRaisesMessage(Exception, "Oops"):
            make_reporter()
        self.assertSequenceEqual(Reporter.objects.all(), [])

    def test_commit(self):
        with transaction.atomic():
            reporter = Reporter.objects.create(first_name="Tintin")
        self.assertSequenceEqual(Reporter.objects.all(), [reporter])

    def test_rollback(self):
        with self.assertRaisesMessage(Exception, "Oops"), transaction.atomic():
            Reporter.objects.create(first_name="Haddock")
            raise Exception("Oops, that's his last name")
        self.assertSequenceEqual(Reporter.objects.all(), [])

    def test_nested_commit_commit(self):
        with transaction.atomic():
            reporter1 = Reporter.objects.create(first_name="Tintin")
            with transaction.atomic():
                reporter2 = Reporter.objects.create(first_name="Archibald", last_name="Haddock")
        self.assertSequenceEqual(Reporter.objects.all(), [reporter2, reporter1])

    def test_nested_rollback_commit(self):
        with self.assertRaisesMessage(Exception, "Oops"), transaction.atomic():
            Reporter.objects.create(last_name="Tintin")
            with transaction.atomic():
                Reporter.objects.create(last_name="Haddock")
            raise Exception("Oops, that's his first name")
        self.assertSequenceEqual(Reporter.objects.all(), [])

    def test_nested_rollback_rollback(self):
        with self.assertRaisesMessage(Exception, "Oops"), transaction.atomic():
            Reporter.objects.create(last_name="Tintin")
            with self.assertRaisesMessage(Exception, "Oops"):
                with transaction.atomic():
                    Reporter.objects.create(first_name="Haddock")
                raise Exception("Oops, that's his last name")
            raise Exception("Oops, that's his first name")
        self.assertSequenceEqual(Reporter.objects.all(), [])

    def test_reuse_commit_commit(self):
        atomic = transaction.atomic()
        with atomic:
            reporter1 = Reporter.objects.create(first_name="Tintin")
            with atomic:
                reporter2 = Reporter.objects.create(first_name="Archibald", last_name="Haddock")
        self.assertSequenceEqual(Reporter.objects.all(), [reporter2, reporter1])

    def test_reuse_rollback_commit(self):
        atomic = transaction.atomic()
        with self.assertRaisesMessage(Exception, "Oops"), atomic:
            Reporter.objects.create(last_name="Tintin")
            with atomic:
                Reporter.objects.create(last_name="Haddock")
            raise Exception("Oops, that's his first name")
        self.assertSequenceEqual(Reporter.objects.all(), [])

    def test_reuse_rollback_rollback(self):
        atomic = transaction.atomic()
        with self.assertRaisesMessage(Exception, "Oops"), atomic:
            Reporter.objects.create(last_name="Tintin")
            with self.assertRaisesMessage(Exception, "Oops"):
                with atomic:
                    Reporter.objects.create(first_name="Haddock")
                raise Exception("Oops, that's his last name")
            raise Exception("Oops, that's his first name")
        self.assertSequenceEqual(Reporter.objects.all(), [])

    def test_rollback_update(self):
        r = Reporter.objects.create(last_name="Tintin")
        with self.assertRaisesMessage(Exception, "Oops"), transaction.atomic():
            Reporter.objects.update(last_name="Haddock")
            # The update is visible in the transaction.
            r.refresh_from_db()
            self.assertEqual(r.last_name, "Haddock")
            raise Exception("Oops")
        # But is now rolled back.
        r.refresh_from_db()
        self.assertEqual(r.last_name, "Tintin")

    def test_rollback_delete(self):
        r = Reporter.objects.create(last_name="Tintin")
        with self.assertRaisesMessage(Exception, "Oops"), transaction.atomic():
            Reporter.objects.all().delete()
            raise Exception("Oops")
        self.assertSequenceEqual(Reporter.objects.all(), [r])

    def test_wrap_callable_instance(self):
        """Atomic can wrap callable instances."""

        class Callable:
            def __call__(self):
                pass

        transaction.atomic(Callable())  # Must not raise an exception

    def test_initializes_connection(self):
        """transaction.atomic() opens the connection if needed."""
        connection.close_pool()
        with transaction.atomic():
            pass


@skipIfDBFeature("_supports_transactions")
class AtomicNotSupportedTests(TransactionTestCase):
    available_apps = ["transactions_"]

    def test_not_supported(self):
        msg = "Transaction numbers are only allowed on a replica set member or mongos"
        with self.assertRaisesMessage(DatabaseError, msg), transaction.atomic():
            Reporter.objects.create(first_name="Haddock")



import copy
import time

import django
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.backends.utils import logger
from django.utils.functional import SimpleLazyObject
from django.utils.text import format_lazy
from django.utils.version import get_version_tuple


def check_django_compatability():
    """
    Verify that this version of django-mongodb-backend is compatible with the
    installed version of Django. For example, any django-mongodb-backend 5.0.x is
    compatible with Django 5.0.y.
    """
    from . import __version__  # noqa: PLC0415

    if django.VERSION[:2] != get_version_tuple(__version__)[:2]:
        A = django.VERSION[0]
        B = django.VERSION[1]
        raise ImproperlyConfigured(
            f"You must use the latest version of django-mongodb-backend {A}.{B}.x "
            f"with Django {A}.{B}.y (found django-mongodb-backend {__version__})."
        )


def prefix_validation_error(error, prefix, code, params):
    """
    Prefix a validation error message while maintaining the existing
    validation data structure.
    """
    if error.error_list == [error]:
        error_params = error.params or {}
        return ValidationError(
            # Messages can't simply be concatenated since they might require
            # their associated parameters to be expressed correctly which is
            # not something format_lazy() does. For example, proxied
            # ngettext calls require a count parameter and are converted
            # to an empty string if they are missing it.
            message=format_lazy(
                "{} {}",
                SimpleLazyObject(lambda: prefix % params),
                SimpleLazyObject(lambda: error.message % error_params),
            ),
            code=code,
            params={**error_params, **params},
        )
    return ValidationError(
        [prefix_validation_error(e, prefix, code, params) for e in error.error_list]
    )


def set_wrapped_methods(cls):
    """Initialize the wrapped methods on cls."""
    if hasattr(cls, "logging_wrapper"):
        for attr in cls.wrapped_methods:
            setattr(cls, attr, cls.logging_wrapper(attr))
        del cls.logging_wrapper
    return cls


@set_wrapped_methods
class OperationDebugWrapper:
    # The PyMongo database and collection methods that this backend uses.
    wrapped_methods = {
        "aggregate",
        "create_collection",
        "create_indexes",
        "create_search_index",
        "drop",
        "find_one",
        "index_information",
        "insert_many",
        "delete_many",
        "drop_index",
        "drop_search_index",
        "list_search_indexes",
        "rename",
        "update_many",
        "update_one",
    }

    def __init__(self, db, collection=None):
        self.collection = collection
        self.db = db
        use_collection = collection is not None
        self.collection_name = f"{collection.name}." if use_collection else ""
        self.wrapped = self.collection if use_collection else self.db.database

    def __getattr__(self, attr):
        return getattr(self.wrapped, attr)

    def profile_call(self, func, args=(), kwargs=None):
        start = time.monotonic()
        retval = func(*args, **kwargs or {})
        duration = time.monotonic() - start
        return duration, retval

    def log(self, op, duration, args, kwargs=None):
        # If kwargs are used by any operations in the future, they must be
        # added to this logging.
        msg = "(%.3f) %s"
        args = ", ".join(repr(arg) for arg in args)
        operation = f"db.{self.collection_name}{op}({args})"
        if len(settings.DATABASES) > 1:
            msg += f"; alias={self.db.alias}"
        self.db.queries_log.append({"sql": operation, "time": f"{duration:.3f}"})
        logger.debug(
            msg,
            duration,
            operation,
            extra={
                "duration": duration,
                "sql": operation,
                "alias": self.db.alias,
            },
        )

    def logging_wrapper(method):
        def wrapper(self, *args, **kwargs):
            func = getattr(self.wrapped, method)
            # Collection.insert_many() mutates args (the documents) by adding
            #  _id. deepcopy() to avoid logging that version.
            original_args = copy.deepcopy(args)
            duration, retval = self.profile_call(func, args, kwargs)
            self.log(method, duration, original_args, kwargs)
            return retval

        return wrapper


@set_wrapped_methods
class OperationCollector(OperationDebugWrapper):
    def __init__(self, collected_sql=None, *, collection=None, db=None):
        super().__init__(db, collection)
        self.collected_sql = collected_sql

    def log(self, op, args, kwargs=None):
        args = ", ".join(repr(arg) for arg in args)
        operation = f"db.{self.collection_name}{op}({args})"
        self.collected_sql.append(operation)

    def logging_wrapper(method):
        def wrapper(self, *args, **kwargs):
            self.log(method, args, kwargs)

        return wrapper


def model_has_encrypted_fields(model):
    """
    Return whether or not this model and its embedded models have any encrypted
    fields.
    """
    from django_mongodb_backend.fields import EmbeddedModelField  # noqa: PLC0415

    return any(
        getattr(field, "encrypted", False)
        or (
            isinstance(field, EmbeddedModelField)
            and model_has_encrypted_fields(field.embedded_model)
        )
        for field in model._meta.fields
    )



from bson import SON
from django.db.models import Sum
from django.test import TestCase

from django_mongodb_backend.test import MongoTestCaseMixin

from .models import Book, Number


class NumericLookupTests(MongoTestCaseMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.objs = Number.objects.bulk_create(Number(num=x) for x in range(5))
        # Null values should be excluded in less than queries.
        cls.null_number = Number.objects.create()

    def test_lt(self):
        self.assertQuerySetEqual(Number.objects.filter(num__lt=3), self.objs[:3])

    def test_lte(self):
        self.assertQuerySetEqual(Number.objects.filter(num__lte=3), self.objs[:4])

    def test_empty_range(self):
        with self.assertNumQueries(0):
            self.assertQuerySetEqual(Number.objects.filter(num__range=[3, 1]), [])

    def test_full_range(self):
        with self.assertNumQueries(1) as ctx:
            self.assertQuerySetEqual(
                Number.objects.filter(num__range=[None, None]), [self.null_number, *self.objs]
            )
        query = ctx.captured_queries[0]["sql"]
        self.assertAggregateQuery(
            query, "lookup__number", [{"$addFields": {"num": "$num"}}, {"$sort": SON([("num", 1)])}]
        )


class RegexTests(MongoTestCaseMixin, TestCase):
    def test_mql(self):
        # $regexMatch must not cast the input to string, otherwise MongoDB
        # can't use the field's indexes.
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(title__regex="Moby Dick"))
        query = ctx.captured_queries[0]["sql"]
        self.assertAggregateQuery(
            query,
            "lookup__book",
            [{"$match": {"title": {"$regex": "Moby Dick", "$options": ""}}}],
        )


class LookupMQLTests(MongoTestCaseMixin, TestCase):
    def test_eq(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(title="Moby Dick"))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"], "lookup__book", [{"$match": {"title": "Moby Dick"}}]
        )

    def test_in(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(title__in=["Moby Dick"]))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "lookup__book",
            [{"$match": {"title": {"$in": ("Moby Dick",)}}}],
        )

    def test_eq_and_in(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(title="Moby Dick", isbn__in=["12345", "56789"]))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "lookup__book",
            [{"$match": {"$and": [{"isbn": {"$in": ("12345", "56789")}}, {"title": "Moby Dick"}]}}],
        )

    def test_gt(self):
        with self.assertNumQueries(1) as ctx:
            list(Number.objects.filter(num__gt=2))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "lookup__number",
            [
                {"$match": {"num": {"$gt": 2}}},
                {"$addFields": {"num": "$num"}},
                {"$sort": SON([("num", 1)])},
            ],
        )

    def test_gte(self):
        with self.assertNumQueries(1) as ctx:
            list(Number.objects.filter(num__gte=2))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "lookup__number",
            [
                {"$match": {"num": {"$gte": 2}}},
                {"$addFields": {"num": "$num"}},
                {"$sort": SON([("num", 1)])},
            ],
        )

    def test_group_by_with_having(self):
        with self.assertNumQueries(1) as ctx:
            list(Number.objects.values("num").annotate(total=Sum("num")).filter(total=1))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "lookup__number",
            [
                {
                    "$group": {
                        "__aggregation1": {"$sum": "$num"},
                        "_id": {"num": "$num"},
                        "total": {"$sum": "$num"},
                    }
                },
                {"$addFields": {"num": "$_id.num"}},
                {"$unset": "_id"},
                {"$match": {"__aggregation1": 1}},
                {"$project": {"num": 1, "total": "$__aggregation1"}},
                {"$sort": SON([("num", 1)])},
            ],
        )

    def test_subquery_filter_constant(self):
        with self.assertNumQueries(1) as ctx:
            list(Number.objects.filter(num__in=Number.objects.filter(num__gt=2).values("num")))
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "lookup__number",
            [
                {
                    "$lookup": {
                        "as": "__subquery0",
                        "from": "lookup__number",
                        "let": {},
                        "pipeline": [
                            {"$match": {"num": {"$gt": 2}}},
                            {"$group": {"_id": None, "subquery_results": {"$addToSet": "$num"}}},
                            {"$addFields": {"__now": "$$NOW"}},
                            {
                                "$unionWith": {
                                    "pipeline": [{"$documents": [{"subquery_results": []}]}]
                                }
                            },
                            {"$limit": 1},
                            {"$project": {"num": "$subquery_results"}},
                        ],
                    }
                },
                {
                    "$set": {
                        "__subquery0": {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": "$__subquery0"}, "missing"]},
                                        {"$eq": [{"$size": "$__subquery0"}, 0]},
                                    ]
                                },
                                "then": {},
                                "else": {"$arrayElemAt": ["$__subquery0", 0]},
                            }
                        }
                    }
                },
                {"$match": {"$expr": {"$in": ["$num", "$__subquery0.num"]}}},
                {"$addFields": {"num": "$num"}},
                {"$sort": SON([("num", 1)])},
            ],
        )



from contextlib import ContextDecorator

from django.db import DEFAULT_DB_ALIAS, DatabaseError
from django.db.transaction import get_connection, on_commit

__all__ = [
    "atomic",
    "on_commit",  # convenience alias
]


class Atomic(ContextDecorator):
    """
    Guarantee the atomic execution of a given block.

    Simplified from django.db.transaction.
    """

    def __init__(self, using):
        self.using = using

    def __enter__(self):
        connection = get_connection(self.using)
        if connection.in_atomic_block_mongo:
            # Track the number of nested atomic() calls.
            connection.nested_atomics += 1
        else:
            # Start a transaction for the outermost atomic().
            connection.start_transaction_mongo()
            connection.in_atomic_block_mongo = True

    def __exit__(self, exc_type, exc_value, traceback):
        connection = get_connection(self.using)
        if connection.nested_atomics:
            # Exiting inner atomic.
            connection.nested_atomics -= 1
        else:
            # Reset flag when exiting outer atomic.
            connection.in_atomic_block_mongo = False
        if exc_type is None:
            # atomic() exited without an error.
            if not connection.in_atomic_block_mongo:
                # Commit transaction if outer atomic().
                try:
                    connection.commit_mongo()
                except DatabaseError:
                    connection.rollback_mongo()
        else:
            # atomic() exited with an error.
            if not connection.in_atomic_block_mongo:
                # Rollback transaction if outer atomic().
                connection.rollback_mongo()


def atomic(using=None):
    # Bare decorator: @atomic -- although the first argument is called `using`,
    # it's actually the function being decorated.
    if callable(using):
        return Atomic(DEFAULT_DB_ALIAS)(using)
    # Decorator: @atomic(...) or context manager: with atomic(...): ...
    return Atomic(using)



from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.db import NotSupportedError, connection
from django.db.backends.signals import connection_created
from django.test import SimpleTestCase, TestCase

from django_mongodb_backend.base import DatabaseWrapper


class DatabaseWrapperTests(SimpleTestCase):
    def test_database_name_empty(self):
        settings = connection.settings_dict.copy()
        settings["NAME"] = ""
        msg = 'settings.DATABASES is missing the "NAME" value.'
        with self.assertRaisesMessage(ImproperlyConfigured, msg):
            DatabaseWrapper(settings)

    def test_database_name_empty_and_host_does_not_contain_database(self):
        settings = connection.settings_dict.copy()
        settings["NAME"] = ""
        settings["HOST"] = "mongodb://localhost"
        msg = 'settings.DATABASES is missing the "NAME" value.'
        with self.assertRaisesMessage(ImproperlyConfigured, msg):
            DatabaseWrapper(settings)

    def test_database_name_parsed_from_host(self):
        settings = connection.settings_dict.copy()
        settings["NAME"] = ""
        settings["HOST"] = "mongodb://localhost/db"
        self.assertEqual(DatabaseWrapper(settings).settings_dict["NAME"], "db")

    def test_database_name_parsed_from_srv_host(self):
        settings = connection.settings_dict.copy()
        settings["NAME"] = ""
        settings["HOST"] = "mongodb+srv://localhost/db"
        # patch() prevents a crash when PyMongo attempts to resolve the
        # nonexistent SRV record.
        with patch("dns.resolver.resolve"):
            self.assertEqual(DatabaseWrapper(settings).settings_dict["NAME"], "db")

    def test_database_name_not_overridden_by_host(self):
        settings = connection.settings_dict.copy()
        settings["NAME"] = "not overridden"
        settings["HOST"] = "mongodb://localhost/db"
        self.assertEqual(DatabaseWrapper(settings).settings_dict["NAME"], "not overridden")


class GetConnectionParamsTests(SimpleTestCase):
    def test_host(self):
        settings = connection.settings_dict.copy()
        settings["HOST"] = "host"
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertEqual(params["host"], "host")

    def test_host_empty(self):
        settings = connection.settings_dict.copy()
        settings["HOST"] = ""
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertIsNone(params["host"])

    def test_user(self):
        settings = connection.settings_dict.copy()
        settings["USER"] = "user"
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertEqual(params["username"], "user")

    def test_password(self):
        settings = connection.settings_dict.copy()
        settings["PASSWORD"] = "password"  # noqa: S105
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertEqual(params["password"], "password")

    def test_port(self):
        settings = connection.settings_dict.copy()
        settings["PORT"] = 123
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertEqual(params["port"], 123)

    def test_port_as_string(self):
        settings = connection.settings_dict.copy()
        settings["PORT"] = "123"
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertEqual(params["port"], 123)

    def test_options(self):
        settings = connection.settings_dict.copy()
        settings["OPTIONS"] = {"extra": "option"}
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertEqual(params["extra"], "option")

    def test_unspecified_settings_omitted(self):
        settings = connection.settings_dict.copy()
        # django.db.utils.ConnectionHandler sets unspecified values to an empty
        # string.
        settings.update(
            {
                "USER": "",
                "PASSWORD": "",
                "PORT": "",
            }
        )
        params = DatabaseWrapper(settings).get_connection_params()
        self.assertNotIn("username", params)
        self.assertNotIn("password", params)
        self.assertNotIn("port", params)


class DatabaseWrapperConnectionTests(TestCase):
    def test_set_autocommit(self):
        self.assertIs(connection.get_autocommit(), True)
        connection.set_autocommit(False)
        self.assertIs(connection.get_autocommit(), False)
        connection.set_autocommit(True)
        self.assertIs(connection.get_autocommit(), True)

    def test_close(self):
        """connection.close() doesn't close the connection."""
        conn = connection.connection
        self.assertIsNotNone(conn)
        connection.close()
        self.assertEqual(connection.connection, conn)

    def test_close_pool(self):
        """connection.close_pool() closes the connection."""
        self.assertIsNotNone(connection.connection)
        connection.close_pool()
        self.assertIsNone(connection.connection)

    def test_connection_created_database_attr(self):
        """
        connection.database is available in the connection_created signal.
        """
        data = {}

        def receiver(sender, connection, **kwargs):  # noqa: ARG001
            data["database"] = connection.database

        connection_created.connect(receiver)
        connection.close_pool()
        # Accessing database implicitly connects.
        connection.database  # noqa: B018
        self.assertIs(data["database"], connection.database)
        connection.close_pool()
        connection_created.disconnect(receiver)
        data.clear()
        connection.connect()
        self.assertEqual(data, {})


class CursorTests(TestCase):
    def test_callproc(self):
        msg = "MongoDB does not support cursor.callproc()."
        with self.assertRaisesMessage(NotSupportedError, msg), connection.cursor() as cursor:
            cursor.callproc("...")
        with self.assertRaisesMessage(NotSupportedError, msg), connection.cursor() as cursor:
            cursor.callproc("...", [])
        with self.assertRaisesMessage(NotSupportedError, msg), connection.cursor() as cursor:
            cursor.callproc("...", [], [])



from django.core.exceptions import EmptyResultSet, FullResultSet
from django.db.models import F
from django.db.models.expressions import CombinedExpression, Func, Value
from django.db.models.sql.query import Query


def is_direct_value(node):
    return not hasattr(node, "as_sql")


def process_lhs(node, compiler, connection, as_expr=False):
    if not hasattr(node, "lhs"):
        # node is a Func or Expression, possibly with multiple source expressions.
        result = []
        for expr in node.get_source_expressions():
            if expr is None:
                continue
            try:
                result.append(expr.as_mql(compiler, connection, as_expr=as_expr))
            except FullResultSet:
                result.append(Value(True).as_mql(compiler, connection, as_expr=as_expr))
            except EmptyResultSet:
                result.append(Value(None).as_mql(compiler, connection, as_expr=as_expr))
        return result
    # node is a Transform with just one source expression, aliased as "lhs".
    if is_direct_value(node.lhs):
        return node
    return node.lhs.as_mql(compiler, connection, as_expr=as_expr)


def process_rhs(node, compiler, connection, as_expr=False):
    rhs = node.rhs
    if hasattr(rhs, "as_mql"):
        if getattr(rhs, "subquery", False) and hasattr(node, "get_subquery_wrapping_pipeline"):
            value = rhs.as_mql(
                compiler,
                connection,
                get_wrapping_pipeline=node.get_subquery_wrapping_pipeline,
                as_expr=as_expr,
            )
        else:
            value = rhs.as_mql(compiler, connection, as_expr=as_expr)
    else:
        _, value = node.process_rhs(compiler, connection)
        # Undo Lookup.get_db_prep_lookup() putting params in a list.
        if not getattr(node, "get_db_prep_lookup_value_is_iterable", False):
            value = value[0]
    return value


def is_constant_value(value):
    if isinstance(value, CombinedExpression):
        # Treat all CombinedExpressions as non-constant until constant cases
        # are handled: https://jira.mongodb.org/browse/INTPYTHON-783.
        return False
    if isinstance(value, list):
        return all(map(is_constant_value, value))
    if is_direct_value(value):
        return True
    if hasattr(value, "get_source_expressions"):
        # Similar limitation as above, sub-expressions should be resolved in
        # the future.
        constants_sub_expressions = all(map(is_constant_value, value.get_source_expressions()))
    else:
        constants_sub_expressions = True
    constants_sub_expressions = constants_sub_expressions and not (
        isinstance(value, Query | F)
        or value.contains_aggregate
        or value.contains_over_clause
        or value.contains_column_references
        or value.contains_subquery
    )
    return constants_sub_expressions and (
        isinstance(value, Value)
        or
        # Some closed functions cannot yet be converted to constant values.
        # Allow Func with can_use_path as a temporary exception.
        (isinstance(value, Func) and value.can_use_path)
    )



import collections


class Adapter(collections.UserDict):
    srid = 4326

    def __init__(self, obj, geography=False):
        """
        Initialize on the spatial object per
        https://www.mongodb.com/docs/manual/reference/geojson/.
        """
        if obj.__class__.__name__ == "GeometryCollection":
            self.data = {
                "type": obj.__class__.__name__,
                "geometries": [self.get_data(x) for x in obj],
            }
        else:
            self.data = self.get_data(obj)

    def get_data(self, obj):
        return {
            "type": obj.__class__.__name__,
            "coordinates": obj.coords,
        }

    @classmethod
    def _fix_polygon(cls, poly):
        return poly



from operator import attrgetter

from django.db import DatabaseError
from django.db.models import Avg, Count, F

from .models import Author, Book, CharModel, DateModel, DateTimeModel, IntegerModel
from .test_base import EncryptionTestCase


class QueryTests(EncryptionTestCase):
    def test_aggregate_avg(self):
        msg = "Accumulator '$avg' cannot aggregate encrypted fields."
        with self.assertRaisesMessage(DatabaseError, msg):
            list(IntegerModel.objects.aggregate(Avg("value")))

    def test_aggregate_count(self):
        msg = "Invalid reference to an encrypted field within aggregate expression: value"
        with self.assertRaisesMessage(DatabaseError, msg):
            list(IntegerModel.objects.aggregate(Count("value")))

    def test_alias(self):
        msg = (
            "Cannot group on field '_id.value' which is encrypted with the "
            "random algorithm or whose encryption properties are not known "
            "until runtime"
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            list(IntegerModel.objects.alias(avg=Avg("value")))

    def test_annotate(self):
        msg = (
            "Cannot group on field '_id.value' which is encrypted with the "
            "random algorithm or whose encryption properties are not known "
            "until runtime"
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            list(IntegerModel.objects.annotate(avg=Avg("value")))

    def test_bulk_create(self):
        CharModel.objects.bulk_create([CharModel(value="abc"), CharModel(value="xyz")])
        self.assertQuerySetEqual(
            CharModel.objects.order_by("pk"), ["abc", "xyz"], attrgetter("value")
        )

    def test_bulk_update(self):
        objs = [
            CharModel.objects.create(value="abc"),
            CharModel.objects.create(value="xyz"),
        ]
        objs[0].value = "def"
        objs[1].value = "mno"
        msg = "Expressions in update queries are not allowed with Queryable Encryption."
        with self.assertRaisesMessage(DatabaseError, msg):
            CharModel.objects.bulk_update(objs, ["value"])

    def test_contains(self):
        obj = CharModel.objects.create(value="abc")
        self.assertIs(CharModel.objects.contains(obj), True)

    def test_count(self):
        CharModel.objects.create(value="a")
        CharModel.objects.create(value="b")
        self.assertEqual(CharModel.objects.count(), 2)

    def test_dates(self):
        msg = (
            "If the value type is a date, the type of the index must also be date (and vice versa)."
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            list(DateModel.objects.dates("value", "year"))

    def test_datetimes(self):
        msg = (
            "If the value type is a date, the type of the index must also be date (and vice versa)."
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            list(DateTimeModel.objects.datetimes("value", "year"))

    def test_distinct(self):
        msg = (
            "Cannot group on field '_id.value' which is encrypted with the "
            "random algorithm or whose encryption properties are not known "
            "until runtime"
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            list(CharModel.objects.distinct("value"))

    def test_exclude(self):
        obj1 = CharModel.objects.create(value="abc")
        obj2 = CharModel.objects.create(value="xyz")
        self.assertSequenceEqual(CharModel.objects.exclude(value=obj1.value), [obj2])

    def test_exists(self):
        self.assertIs(CharModel.objects.exists(), False)

    def test_get_or_create(self):
        obj1, created1 = CharModel.objects.get_or_create(value="abc")
        self.assertIs(created1, True)
        obj2, created2 = CharModel.objects.get_or_create(value="abc")
        self.assertIs(created2, False)
        self.assertEqual(obj1, obj2)

    def test_join(self):
        book = Book.objects.create(title="Book", author=Author.objects.create(name="Bob"))
        self.assertSequenceEqual(Book.objects.filter(author__name="Bob"), [book])

    def test_join_with_let(self):
        msg = (
            "Non-empty 'let' field is not allowed in the $lookup aggregation "
            "stage over an encrypted collection."
        )
        with self.assertRaisesMessage(DatabaseError, msg):
            list(Book.objects.filter(author__name=F("title")))

    def test_order_by(self):
        msg = "Cannot add an encrypted field as a prefix of another encrypted field"
        with self.assertRaisesMessage(DatabaseError, msg):
            list(CharModel.objects.order_by("value"))

    def test_select_related(self):
        Book.objects.create(title="Book", author=Author.objects.create(name="Bob"))
        with self.assertNumQueries(1, using="encrypted"):
            books = Book.objects.select_related("author")
            self.assertEqual(books[0].author.name, "Bob")

    def test_update(self):
        msg = "Multi-document updates are not allowed with Queryable Encryption"
        with self.assertRaisesMessage(DatabaseError, msg):
            self.assertEqual(CharModel.objects.update(value="xyz"), 1)

    def test_update_or_create(self):
        obj, created = CharModel.objects.update_or_create(value="xyz")
        self.assertIs(created, True)
        self.assertEqual(obj.value, "xyz")
        obj, created = CharModel.objects.update_or_create(value="xyz", defaults={"plain": "abc"})
        self.assertIs(created, False)
        self.assertEqual(obj.plain, "abc")

    def test_union(self):
        msg = "Aggregation stage $unionWith is not allowed or supported with automatic encryption."
        qs1 = IntegerModel.objects.filter(value__gt=1)
        qs2 = IntegerModel.objects.filter(value__gte=8)
        with self.assertRaisesMessage(DatabaseError, msg):
            list(qs1.union(qs2))

    def test_values(self):
        list(CharModel.objects.values("value"))

    def test_values_list(self):
        list(CharModel.objects.values_list("value"))



import datetime
from decimal import Decimal
from functools import partialmethod
from uuid import UUID

from bson import Decimal128
from django.core.exceptions import EmptyResultSet, FullResultSet
from django.db import NotSupportedError
from django.db.models.expressions import (
    BaseExpression,
    Case,
    Col,
    ColPairs,
    CombinedExpression,
    Exists,
    ExpressionList,
    ExpressionWrapper,
    NegatedExpression,
    OrderBy,
    RawSQL,
    Ref,
    ResolvedOuterRef,
    Star,
    Subquery,
    Value,
    When,
)
from django.db.models.sql import Query

from django_mongodb_backend.query_utils import process_lhs


def base_expression(self, compiler, connection, as_expr=False, **extra):
    # Use as_mql_path(), if possible.
    if not as_expr and hasattr(self, "as_mql_path") and getattr(self, "can_use_path", False):
        return self.as_mql_path(compiler, connection, **extra)
    # Otherwise, use as_mql_expr().
    expr = self.as_mql_expr(compiler, connection, **extra)
    return expr if as_expr else {"$expr": expr}


def case(self, compiler, connection):
    case_parts = []
    for case in self.cases:
        case_mql = {}
        try:
            case_mql["case"] = case.as_mql(compiler, connection, as_expr=True)
        except EmptyResultSet:
            continue
        except FullResultSet:
            default_mql = case.result.as_mql(compiler, connection, as_expr=True)
            break
        case_mql["then"] = case.result.as_mql(compiler, connection, as_expr=True)
        case_parts.append(case_mql)
    else:
        default_mql = self.default.as_mql(compiler, connection, as_expr=True)
    if not case_parts:
        return default_mql
    return {
        "$switch": {
            "branches": case_parts,
            "default": default_mql,
        }
    }


def col(self, compiler, connection, as_expr=False):  # noqa: ARG001
    # If the column is part of a subquery and belongs to one of the parent
    # queries, it will be stored for reference using $let in a $lookup stage.
    # If the query is built with `alias_cols=False`, treat the column as
    # belonging to the current collection.
    if self.alias is not None and (
        self.alias not in compiler.query.alias_refcount
        or compiler.query.alias_refcount[self.alias] == 0
    ):
        try:
            index = compiler.column_indices[self]
        except KeyError:
            index = len(compiler.column_indices)
            compiler.column_indices[self] = index
        return f"$${compiler.PARENT_FIELD_TEMPLATE.format(index)}"
    # Add the column's collection's alias for columns in joined collections.
    has_alias = self.alias and self.alias != compiler.collection_name
    prefix = f"{self.alias}." if has_alias else ""
    if as_expr:
        prefix = f"${prefix}"
    return f"{prefix}{self.target.column}"


def col_pairs(self, compiler, connection, as_expr=False):
    cols = self.get_cols()
    if len(cols) > 1:
        raise NotSupportedError("ColPairs is not supported.")
    return cols[0].as_mql(compiler, connection, as_expr=as_expr)


def combined_expression(self, compiler, connection):
    expressions = [
        self.lhs.as_mql(compiler, connection, as_expr=True),
        self.rhs.as_mql(compiler, connection, as_expr=True),
    ]
    return connection.ops.combine_expression(self.connector, expressions)


def expression_wrapper(self, compiler, connection):
    return self.expression.as_mql(compiler, connection, as_expr=True)


def negated_expression(self, compiler, connection):
    return {"$not": expression_wrapper(self, compiler, connection)}


def order_by(self, compiler, connection, as_expr=False):
    return self.expression.as_mql(compiler, connection, as_expr=as_expr)


@property
def order_by_can_use_path(self):
    return self.expression.is_simple_column


def query(self, compiler, connection, get_wrapping_pipeline=None, as_expr=False):
    subquery_compiler = self.get_compiler(connection=connection)
    subquery_compiler.pre_sql_setup(with_col_aliases=False)
    field_name, expr = subquery_compiler.columns[0]
    subquery = subquery_compiler.build_query(
        subquery_compiler.columns
        if subquery_compiler.query.annotations or not subquery_compiler.query.default_cols
        else None
    )
    table_output = f"__subquery{len(compiler.subqueries)}"
    from_table = next(
        e.table_name for alias, e in self.alias_map.items() if self.alias_refcount[alias]
    )
    # To perform a subquery, a $lookup stage that escapsulates the entire
    # subquery pipeline is added. The "let" clause defines the variables
    # needed to bridge the main collection with the subquery.
    subquery.subquery_lookup = {
        "as": table_output,
        "from": from_table,
        "let": {
            compiler.PARENT_FIELD_TEMPLATE.format(i): col.as_mql(compiler, connection, as_expr=True)
            for col, i in subquery_compiler.column_indices.items()
        },
    }
    if get_wrapping_pipeline:
        # The results from some lookups must be converted to a list of values.
        # The output is compressed with an aggregation pipeline.
        wrapping_result_pipeline = get_wrapping_pipeline(
            subquery_compiler, connection, field_name, expr
        )
        # If the subquery is a combinator, wrap the result at the end of the
        # combinator pipeline...
        if subquery.query.combinator:
            subquery.combinator_pipeline.extend(wrapping_result_pipeline)
        # ... otherwise put at the end of subquery's pipeline.
        else:
            if subquery.aggregation_pipeline is None:
                subquery.aggregation_pipeline = []
            subquery.aggregation_pipeline.extend(wrapping_result_pipeline)
        # Erase project_fields since the required value is projected above.
        subquery.project_fields = None
    compiler.subqueries.append(subquery)
    if as_expr:
        return f"${table_output}.{field_name}"
    return f"{table_output}.{field_name}"


def raw_sql(self, compiler, connection, as_expr=False):  # noqa: ARG001
    raise NotSupportedError("RawSQL is not supported on MongoDB.")


def ref(self, compiler, connection, as_expr=False):  # noqa: ARG001
    prefix = (
        f"{self.source.alias}."
        if isinstance(self.source, Col) and self.source.alias != compiler.collection_name
        else ""
    )
    if hasattr(self, "ordinal"):
        refs, _ = compiler.columns[self.ordinal - 1]
    else:
        refs = self.refs
    if as_expr:
        prefix = f"${prefix}"
    return f"{prefix}{refs}"


def star(self, compiler, connection):  # noqa: ARG001
    return {"$literal": True}


def subquery(self, compiler, connection, get_wrapping_pipeline=None, as_expr=False):
    return self.query.as_mql(
        compiler, connection, get_wrapping_pipeline=get_wrapping_pipeline, as_expr=as_expr
    )


def exists(self, compiler, connection, get_wrapping_pipeline=None):
    try:
        lhs_mql = subquery(
            self, compiler, connection, get_wrapping_pipeline=get_wrapping_pipeline, as_expr=True
        )
    except EmptyResultSet:
        return Value(False).as_mql(compiler, connection, as_expr=True)
    return connection.mongo_expr_operators["isnull"](lhs_mql, False)


def when(self, compiler, connection):
    return self.condition.as_mql(compiler, connection, as_expr=True)


def value(self, compiler, connection, as_expr=False):  # noqa: ARG001
    value = self.value
    if isinstance(value, (list, int, str, dict, tuple)) and as_expr:
        # Wrap lists, numbers, strings, dicts, and tuples in $literal to avoid
        # ambiguity when Value is used in aggregate() or update_many()'s $set.
        return {"$literal": value}
    if isinstance(value, Decimal):
        return Decimal128(value)
    if isinstance(value, datetime.datetime):
        return value
    if isinstance(value, datetime.date):
        # Turn dates into datetimes since BSON doesn't support dates.
        return datetime.datetime.combine(value, datetime.datetime.min.time())
    if isinstance(value, datetime.time):
        # Turn times into datetimes since BSON doesn't support times.
        return datetime.datetime.combine(datetime.datetime.min.date(), value)
    if isinstance(value, datetime.timedelta):
        # DurationField stores milliseconds rather than microseconds.
        return value / datetime.timedelta(milliseconds=1)
    if isinstance(value, UUID):
        return value.hex
    return value


def register_expressions():
    BaseExpression.as_mql = base_expression
    BaseExpression.is_simple_column = False
    Case.as_mql_expr = case
    Col.as_mql = col
    Col.is_simple_column = True
    ColPairs.as_mql = col_pairs
    CombinedExpression.as_mql_expr = combined_expression
    Exists.as_mql_expr = exists
    ExpressionList.as_mql = process_lhs
    ExpressionWrapper.as_mql_expr = expression_wrapper
    NegatedExpression.as_mql_expr = negated_expression
    OrderBy.as_mql_expr = partialmethod(order_by, as_expr=True)
    OrderBy.as_mql_path = partialmethod(order_by, as_expr=False)
    OrderBy.can_use_path = order_by_can_use_path
    Query.as_mql = query
    RawSQL.as_mql = raw_sql
    Ref.as_mql = ref
    Ref.is_simple_column = True
    ResolvedOuterRef.as_mql = ResolvedOuterRef.as_sql
    Star.as_mql_expr = star
    Subquery.as_mql_expr = partialmethod(subquery, as_expr=True)
    Subquery.as_mql_path = partialmethod(subquery, as_expr=False)
    When.as_mql_expr = when
    Value.as_mql = value



# Settings for django/tests.
import os

from pymongo.uri_parser import parse_uri

if mongodb_uri := os.getenv("MONGODB_URI"):
    db_settings = {
        "ENGINE": "django_mongodb_backend",
        "HOST": mongodb_uri,
    }
    # Workaround for https://github.com/mongodb-labs/mongo-orchestration/issues/268
    uri = parse_uri(mongodb_uri)
    if uri.get("username") and uri.get("password"):
        db_settings["OPTIONS"] = {"tls": True, "tlsAllowInvalidCertificates": True}
    DATABASES = {
        "default": {**db_settings, "NAME": "djangotests"},
        "other": {**db_settings, "NAME": "djangotests-other"},
    }
else:
    DATABASES = {
        "default": {
            "ENGINE": "django_mongodb_backend",
            "NAME": "djangotests",
            # Required when connecting to the Atlas image in Docker.
            "OPTIONS": {"directConnection": True},
        },
        "other": {
            "ENGINE": "django_mongodb_backend",
            "NAME": "djangotests-other",
            "OPTIONS": {"directConnection": True},
        },
    }

DEFAULT_AUTO_FIELD = "django_mongodb_backend.fields.ObjectIdAutoField"
PASSWORD_HASHERS = ("django.contrib.auth.hashers.MD5PasswordHasher",)
SECRET_KEY = "django_tests_secret_key"  # noqa: S105
USE_TZ = False



import functools
import sys
import types

from django.conf import settings
from django.db.backends.base.creation import TEST_DATABASE_PREFIX, BaseDatabaseCreation
from django.utils.module_loading import import_string


def assertRaises(exception, message):
    """
    Wrap a test to assert that it raise the given exception with the given
    message.
    """

    def decorator(test):
        # Protect against a class reference in django_test_expected_raises.
        if not isinstance(test, types.FunctionType):
            raise TypeError(
                "Items in django_test_expected_raises must be test methods (got {test})."
            )

        @functools.wraps(test)
        def assert_raises_wrapper(self, *args, **kwargs):
            with self.assertRaisesMessage(exception, message):
                test(self, *args, **kwargs)

        return assert_raises_wrapper

    return decorator


class DatabaseCreation(BaseDatabaseCreation):
    def _execute_create_test_db(self, cursor, parameters, keepdb=False):
        # Close the connection (which may point to the non-test database) so
        # that a new connection to the test database can be established later.
        self.connection.close_pool()
        # Use a test _key_vault_namespace. This assumes the key vault database
        # is the same as the encrypted database so that _destroy_test_db() can
        # reset the collection by dropping it.
        if opts := self.connection.settings_dict["OPTIONS"].get("auto_encryption_opts"):
            self.connection.settings_dict["OPTIONS"][
                "auto_encryption_opts"
            ]._key_vault_namespace = TEST_DATABASE_PREFIX + opts._key_vault_namespace
        if not keepdb:
            self._destroy_test_db(parameters["dbname"], verbosity=0)

    def _destroy_test_db(self, test_database_name, verbosity):
        # At this point, settings still points to the non-test database. For
        # MongoDB, it must use the test database.
        settings.DATABASES[self.connection.alias]["NAME"] = test_database_name
        self.connection.settings_dict["NAME"] = test_database_name

        for collection in self.connection.introspection.table_names():
            if not collection.startswith("system."):
                self.connection.database.drop_collection(collection)

    def destroy_test_db(self, old_database_name=None, verbosity=1, keepdb=False, suffix=None):
        super().destroy_test_db(old_database_name, verbosity, keepdb, suffix)
        # Close the connection to the test database.
        self.connection.close_pool()
        # Restore the original _key_vault_namespace.
        if opts := self.connection.settings_dict["OPTIONS"].get("auto_encryption_opts"):
            self.connection.settings_dict["OPTIONS"][
                "auto_encryption_opts"
            ]._key_vault_namespace = opts._key_vault_namespace[len(TEST_DATABASE_PREFIX) :]

    def mark_expected_failures_and_skips(self):
        super().mark_expected_failures_and_skips()
        # Add an assertion wrapper to tests that are expected to raise an
        # exception (most often NotSupportedError).
        if self.connection.alias != "default":
            # Wrap tests only once since assertRaises() doesn't work if nested.
            return
        for (exception, msg), tests in self.connection.features.django_test_expected_raises.items():
            for test_name in tests:
                module_or_class_name, _, name_to_mark = test_name.rpartition(".")
                test_app = test_name.split(".")[0]
                # Importing a test app that isn't installed raises RuntimeError.
                if test_app in settings.INSTALLED_APPS:
                    try:
                        test_frame = import_string(module_or_class_name)
                    except ImportError:
                        # Same comment from similar logic in superclass.
                        test_to_mark = import_string(test_name)
                        test_frame = sys.modules.get(test_to_mark.__module__)
                    else:
                        test_to_mark = getattr(test_frame, name_to_mark)
                    setattr(test_frame, name_to_mark, assertRaises(exception, msg)(test_to_mark))



from django.test import TestCase

from django_mongodb_backend.forms import EmbeddedModelArrayField

from .forms import MovieForm, StoreForm
from .models import Movie, Product, Review, Store


class ModelFormTests(TestCase):
    def test_add_another(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Great!",
            "reviews-0-rating": "10",
            "reviews-1-title": "Not so great",
            "reviews-1-rating": "1",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(form.changed_data, ["reviews"])
        movie.refresh_from_db()
        self.assertEqual(len(movie.reviews), 2)
        review = movie.reviews[0]
        self.assertEqual(review.title, "Great!")
        self.assertEqual(review.rating, 10)
        review = movie.reviews[1]
        self.assertEqual(review.title, "Not so great")
        self.assertEqual(review.rating, 1)

    def test_no_change(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Great!",
            "reviews-0-rating": "10",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(form.changed_data, [])
        movie.refresh_from_db()
        self.assertEqual(len(movie.reviews), 1)
        review = movie.reviews[0]
        self.assertEqual(review.title, "Great!")
        self.assertEqual(review.rating, 10)

    def test_update(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Not so great",
            "reviews-0-rating": "1",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertTrue(form.is_valid())
        form.save()
        self.assertEqual(form.changed_data, ["reviews"])
        movie.refresh_from_db()
        self.assertEqual(len(movie.reviews), 1)
        review = movie.reviews[0]
        self.assertEqual(review.title, "Not so great")
        self.assertEqual(review.rating, 1)

    def test_some_missing_data(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "",
            "reviews-0-rating": "1",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["reviews"], ["This field is required."])

    def test_invalid_field_data(self):
        """A field's data (rating) is invalid."""
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Great!",
            "reviews-0-rating": "not a number",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["reviews"],
            ["Enter a whole number."],
        )

    def test_all_missing_data(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "",
            "reviews-0-rating": "",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["reviews"], ["This field is required.", "This field is required."]
        )

    def test_delete(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10), Review(title="Okay", rating=5)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Not so great",
            "reviews-0-rating": "1",
            "reviews-0-DELETE": "1",
            "reviews-1-title": "Okay",
            "reviews-1-rating": "5",
            "reviews-1-DELETE": "",
            "reviews-TOTAL_FORMS": 3,
            "reviews-INITIAL_FORMS": 2,
        }
        form = MovieForm(data, instance=movie)
        self.assertTrue(form.is_valid())
        form.save()
        movie.refresh_from_db()
        self.assertEqual(len(movie.reviews), 1)
        review = movie.reviews[0]
        self.assertEqual(review.title, "Okay")
        self.assertEqual(review.rating, 5)

    def test_delete_required(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Not so great",
            "reviews-0-rating": "1",
            "reviews-0-DELETE": "1",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["reviews"], ["This field cannot be blank."])

    def test_max_size(self):
        """
        Submitting more than the allowed number of items (three featured
        reviews for max_size=2) is prohibited.
        """
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
            featured_reviews=[Review(title="Okay", rating=5)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Not so great",
            "reviews-0-rating": "1",
            "reviews-0-DELETE": "",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
            "featured_reviews-0-title": "Okay",
            "featured_reviews-0-rating": "5",
            "featured_reviews-1-title": "Okay",
            "featured_reviews-1-rating": "5",
            "featured_reviews-2-title": "Okay",
            "featured_reviews-2-rating": "5",
            "featured_reviews-TOTAL_FORMS": 3,
            "featured_reviews-INITIAL_FORMS": 0,
        }
        form = MovieForm(data, instance=movie)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["featured_reviews"], ["Please submit at most 2 forms."])

    def test_nullable_field(self):
        """A nullable field is emptied if all rows are deleted."""
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
            featured_reviews=[Review(title="Okay", rating=5)],
        )
        data = {
            "title": "Lion King",
            "reviews-0-title": "Not so great",
            "reviews-0-rating": "1",
            "reviews-0-DELETE": "",
            "reviews-TOTAL_FORMS": 2,
            "reviews-INITIAL_FORMS": 1,
            "featured_reviews-0-title": "Okay",
            "featured_reviews-0-rating": "5",
            "featured_reviews-0-DELETE": "1",
            "featured_reviews-TOTAL_FORMS": 2,
            "featured_reviews-INITIAL_FORMS": 1,
        }
        form = MovieForm(data, instance=movie)
        self.assertTrue(form.is_valid())
        form.save()
        movie.refresh_from_db()
        self.assertEqual(len(movie.featured_reviews), 0)

    def test_rendering(self):
        form = MovieForm()
        self.assertHTMLEqual(
            str(form.fields["reviews"].get_bound_field(form, "reviews").label_tag()),
            '<label for="id_reviews">Reviews:</label>',
        )
        self.assertHTMLEqual(
            str(form.fields["reviews"].get_bound_field(form, "reviews")),
            """
<table>
<tbody>
  <tr>
    <th><label for="id_reviews-0-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-0-title" maxlength="255" id="id_reviews-0-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-0-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-0-rating" id="id_reviews-0-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-0-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-0-DELETE" id="id_reviews-0-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_reviews-1-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-1-title" maxlength="255" id="id_reviews-1-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-1-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-1-rating" id="id_reviews-1-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-1-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-1-DELETE" id="id_reviews-1-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_reviews-2-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-2-title" maxlength="255" id="id_reviews-2-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-2-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-2-rating" id="id_reviews-2-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-2-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-2-DELETE" id="id_reviews-2-DELETE">
    </td>
  </tr>
</tbody>
</table>
<input type="hidden" name="reviews-TOTAL_FORMS" value="3"
    id="id_reviews-TOTAL_FORMS"><input type="hidden"
    name="reviews-INITIAL_FORMS" value="0"
    id="id_reviews-INITIAL_FORMS">
<input type="hidden" name="reviews-MIN_NUM_FORMS" value="0"
    id="id_reviews-MIN_NUM_FORMS"><input type="hidden"
    name="reviews-MAX_NUM_FORMS" value="1000" id="id_reviews-MAX_NUM_FORMS">""",
        )

    def test_rendering_initial(self):
        movie = Movie.objects.create(
            title="Lion King",
            reviews=[Review(title="Great!", rating=10)],
        )
        form = MovieForm(instance=movie)
        self.assertHTMLEqual(
            str(form.fields["reviews"].get_bound_field(form, "reviews")),
            """
<table>
<tbody>
  <tr>
    <th><label for="id_reviews-0-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-0-title" maxlength="255"
        id="id_reviews-0-title" value="Great!">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-0-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-0-rating"
        id="id_reviews-0-rating" value="10">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-0-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-0-DELETE" id="id_reviews-0-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_reviews-1-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-1-title" maxlength="255" id="id_reviews-1-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-1-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-1-rating" id="id_reviews-1-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-1-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-1-DELETE" id="id_reviews-1-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_reviews-2-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-2-title" maxlength="255" id="id_reviews-2-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-2-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-2-rating" id="id_reviews-2-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-2-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-2-DELETE" id="id_reviews-2-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_reviews-3-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-3-title" maxlength="255" id="id_reviews-3-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-3-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-3-rating" id="id_reviews-3-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-3-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-3-DELETE" id="id_reviews-3-DELETE">
    </td>
  </tr>
</tbody>
</table>
<input type="hidden" name="reviews-TOTAL_FORMS" value="4"
    id="id_reviews-TOTAL_FORMS"><input type="hidden"
    name="reviews-INITIAL_FORMS" value="1"
    id="id_reviews-INITIAL_FORMS">
<input type="hidden" name="reviews-MIN_NUM_FORMS" value="0"
    id="id_reviews-MIN_NUM_FORMS"><input type="hidden"
    name="reviews-MAX_NUM_FORMS" value="1000" id="id_reviews-MAX_NUM_FORMS">""",
        )

    def test_extra_forms(self):
        """The extra_forms argument specifies the number of extra forms."""

        class ExtraMovieForm(MovieForm):
            reviews = EmbeddedModelArrayField(Review, prefix="reviews", extra_forms=2)

        form = ExtraMovieForm()
        self.assertHTMLEqual(
            str(form.fields["reviews"].get_bound_field(form, "reviews")),
            """
<table>
<tbody>
  <tr>
    <th><label for="id_reviews-0-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-0-title" maxlength="255"
        id="id_reviews-0-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-0-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-0-rating" id="id_reviews-0-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-0-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-0-DELETE" id="id_reviews-0-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_reviews-1-title">Title:</label></th>
    <td>
      <input type="text" name="reviews-1-title" maxlength="255" id="id_reviews-1-title">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-1-rating">Rating:</label></th>
    <td>
      <input type="number" name="reviews-1-rating" id="id_reviews-1-rating">
    </td>
  </tr>
  <tr>
    <th><label for="id_reviews-1-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="reviews-1-DELETE" id="id_reviews-1-DELETE">
    </td>
  </tr>
</tbody>
</table>
<input type="hidden" name="reviews-TOTAL_FORMS" value="2"
    id="id_reviews-TOTAL_FORMS"><input type="hidden"
    name="reviews-INITIAL_FORMS" value="0"
    id="id_reviews-INITIAL_FORMS">
<input type="hidden" name="reviews-MIN_NUM_FORMS" value="0"
    id="id_reviews-MIN_NUM_FORMS"><input type="hidden"
    name="reviews-MAX_NUM_FORMS" value="1000" id="id_reviews-MAX_NUM_FORMS">""",
        )


class NestedFormTests(TestCase):
    def test_update(self):
        store = Store.objects.create(
            name="Best Buy",
            products=[Product(name="TV", reviews=[Review(title="Great", rating=9)])],
        )
        data = {
            "name": "Best Buy!",
            "products-0-name": "TV!",
            "products-0-reviews-0-title": "Great!",
            "products-0-reviews-0-rating": "9",
            "products-TOTAL_FORMS": 3,
            "products-INITIAL_FORMS": 1,
            "products-0-reviews-TOTAL_FORMS": 3,
            "products-0-reviews-INITIAL_FORMS": 1,
        }
        form = StoreForm(data, instance=store)
        self.assertTrue(form.is_valid())
        form.save()
        store.refresh_from_db()
        self.assertEqual(store.name, "Best Buy!")
        self.assertEqual(store.products[0].name, "TV!")
        self.assertEqual(store.products[0].reviews[0].title, "Great!")

    def test_delete(self):
        """The "Bad" review is deleted."""
        store = Store.objects.create(
            name="Best Buy",
            products=[
                Product(
                    name="TV",
                    reviews=[
                        Review(title="Great", rating=9),
                        Review(title="Bad", rating=1),
                    ],
                )
            ],
        )
        data = {
            "name": "Best Buy",
            "products-0-name": "TV",
            "products-0-reviews-0-title": "Great",
            "products-0-reviews-0-rating": "9",
            "products-0-reviews-1-title": "Bad",
            "products-0-reviews-1-rating": "1",
            "products-0-reviews-1-DELETE": "1",
            "products-TOTAL_FORMS": 3,
            "products-INITIAL_FORMS": 1,
            "products-0-reviews-TOTAL_FORMS": 3,
            "products-0-reviews-INITIAL_FORMS": 2,
        }
        form = StoreForm(data, instance=store)
        self.assertTrue(form.is_valid())
        form.save()
        store.refresh_from_db()
        self.assertEqual(len(store.products[0].reviews), 1)
        self.assertEqual(store.products[0].reviews[0].title, "Great")

    def test_some_missing_data(self):
        """A required field (Review.title) is missing."""
        store = Store.objects.create(
            name="Best Buy",
            products=[Product(name="TV", reviews=[Review(title="Great", rating=9)])],
        )
        data = {
            "name": "Best Buy!",
            "products-0-name": "TV!",
            "products-0-reviews-0-title": "",
            "products-0-reviews-0-rating": "9",
            "products-TOTAL_FORMS": 3,
            "products-INITIAL_FORMS": 1,
            "products-0-reviews-TOTAL_FORMS": 3,
            "products-0-reviews-INITIAL_FORMS": 1,
            "products-1-reviews-TOTAL_FORMS": 3,
            "products-1-reviews-INITIAL_FORMS": 0,
            "products-2-reviews-TOTAL_FORMS": 3,
            "products-2-reviews-INITIAL_FORMS": 0,
        }
        form = StoreForm(data, instance=store)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["products"], ["This field is required."])
        self.assertHTMLEqual(
            str(form),
            """
<div>
  <label for="id_name">Name:</label>
  <input type="text" name="name" value="Best Buy!" maxlength="255" required id="id_name">
</div>
<div>
  <label for="id_products">Products:</label>
  <ul class="errorlist" id="id_products_error">
    <li>This field is required.</li>
  </ul>
  <table>
    <tbody>
      <tr>
        <th><label for="id_products-0-name">Name:</label></th>
        <td>
          <input type="text" name="products-0-name" value="TV!" maxlength="255"
           id="id_products-0-name">
        </td>
      </tr>
      <tr>
        <th><label for="id_products-0-reviews">Reviews:</label></th>
        <td>
          <ul class="errorlist" id="id_products-0-reviews_error">
            <li>This field is required.</li>
          </ul>
          <table>
            <tbody>
              <tr>
                <th><label for="id_products-0-reviews-0-title">Title:</label></th>
                <td>
                  <ul class="errorlist" id="id_products-0-reviews-0-title_error">
                    <li>This field is required.</li>
                  </ul>
                  <input type="text" name="products-0-reviews-0-title" maxlength="255"
                   aria-invalid="true" aria-describedby="id_products-0-reviews-0-title_error"
                   id="id_products-0-reviews-0-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-0-reviews-0-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-0-reviews-0-rating" value="9"
                   id="id_products-0-reviews-0-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-0-reviews-0-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-0-reviews-0-DELETE"
                   id="id_products-0-reviews-0-DELETE">
                </td>
              </tr>
            </tbody>
            <tbody>
              <tr>
                <th><label for="id_products-0-reviews-1-title">Title:</label></th>
                <td>
                  <input type="text" name="products-0-reviews-1-title" maxlength="255"
                   id="id_products-0-reviews-1-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-0-reviews-1-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-0-reviews-1-rating"
                   id="id_products-0-reviews-1-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-0-reviews-1-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-0-reviews-1-DELETE"
                   id="id_products-0-reviews-1-DELETE">
                </td>
              </tr>
            </tbody>
            <tbody>
              <tr>
                <th><label for="id_products-0-reviews-2-title">Title:</label></th>
                <td>
                  <input type="text" name="products-0-reviews-2-title" maxlength="255"
                   id="id_products-0-reviews-2-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-0-reviews-2-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-0-reviews-2-rating"
                   id="id_products-0-reviews-2-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-0-reviews-2-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-0-reviews-2-DELETE"
                   id="id_products-0-reviews-2-DELETE">
                </td>
              </tr>
            </tbody>
          </table>
          <input type="hidden" name="products-0-reviews-TOTAL_FORMS" value="3"
           id="id_products-0-reviews-TOTAL_FORMS">
          <input type="hidden" name="products-0-reviews-INITIAL_FORMS" value="1"
            id="id_products-0-reviews-INITIAL_FORMS">
          <input type="hidden" name="products-0-reviews-MIN_NUM_FORMS"
           id="id_products-0-reviews-MIN_NUM_FORMS">
          <input type="hidden" name="products-0-reviews-MAX_NUM_FORMS"
           id="id_products-0-reviews-MAX_NUM_FORMS">
        </td>
      </tr>
      <tr>
        <th><label for="id_products-0-DELETE">Delete:</label></th>
        <td>
          <input type="checkbox" name="products-0-DELETE" id="id_products-0-DELETE">
        </td>
      </tr>
    </tbody>
    <tbody>
      <tr>
        <th><label for="id_products-1-name">Name:</label></th>
        <td>
          <input type="text" name="products-1-name" maxlength="255" id="id_products-1-name">
        </td>
      </tr>
      <tr>
        <th><label for="id_products-1-reviews">Reviews:</label></th>
        <td>
          <table>
            <tbody>
              <tr>
                <th><label for="id_products-1-reviews-0-title">Title:</label></th>
                <td>
                  <input type="text" name="products-1-reviews-0-title"
                   maxlength="255" id="id_products-1-reviews-0-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-1-reviews-0-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-1-reviews-0-rating"
                   id="id_products-1-reviews-0-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-1-reviews-0-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-1-reviews-0-DELETE"
                   id="id_products-1-reviews-0-DELETE">
                </td>
              </tr>
            </tbody>
            <tbody>
              <tr>
                <th><label for="id_products-1-reviews-1-title">Title:</label></th>
                <td>
                  <input type="text" name="products-1-reviews-1-title" maxlength="255"
                   id="id_products-1-reviews-1-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-1-reviews-1-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-1-reviews-1-rating"
                   id="id_products-1-reviews-1-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-1-reviews-1-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-1-reviews-1-DELETE"
                   id="id_products-1-reviews-1-DELETE">
                </td>
              </tr>
            </tbody>
            <tbody>
              <tr>
                <th><label for="id_products-1-reviews-2-title">Title:</label></th>
                <td>
                  <input type="text" name="products-1-reviews-2-title" maxlength="255"
                   id="id_products-1-reviews-2-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-1-reviews-2-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-1-reviews-2-rating"
                   id="id_products-1-reviews-2-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-1-reviews-2-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-1-reviews-2-DELETE"
                   id="id_products-1-reviews-2-DELETE">
                </td>
              </tr>
            </tbody>
          </table>
          <input type="hidden" name="products-1-reviews-TOTAL_FORMS"
           value="3" id="id_products-1-reviews-TOTAL_FORMS">
          <input type="hidden" name="products-1-reviews-INITIAL_FORMS"
           value="0" id="id_products-1-reviews-INITIAL_FORMS">
          <input type="hidden" name="products-1-reviews-MIN_NUM_FORMS"
           id="id_products-1-reviews-MIN_NUM_FORMS">
          <input type="hidden" name="products-1-reviews-MAX_NUM_FORMS"
           id="id_products-1-reviews-MAX_NUM_FORMS">
        </td>
      </tr>
      <tr>
        <th><label for="id_products-1-DELETE">Delete:</label></th>
        <td>
          <input type="checkbox" name="products-1-DELETE" id="id_products-1-DELETE">
        </td>
      </tr>
    </tbody>
    <tbody>
      <tr>
        <th><label for="id_products-2-name">Name:</label></th>
        <td>
          <input type="text" name="products-2-name" maxlength="255" id="id_products-2-name">
        </td>
      </tr>
      <tr>
        <th><label for="id_products-2-reviews">Reviews:</label></th>
        <td>
          <table>
            <tbody>
              <tr>
                <th><label for="id_products-2-reviews-0-title">Title:</label></th>
                <td>
                  <input type="text" name="products-2-reviews-0-title" maxlength="255"
                   id="id_products-2-reviews-0-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-2-reviews-0-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-2-reviews-0-rating"
                   id="id_products-2-reviews-0-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-2-reviews-0-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-2-reviews-0-DELETE"
                   id="id_products-2-reviews-0-DELETE">
                </td>
              </tr>
            </tbody>
            <tbody>
              <tr>
                <th><label for="id_products-2-reviews-1-title">Title:</label></th>
                <td>
                  <input type="text" name="products-2-reviews-1-title" maxlength="255"
                   id="id_products-2-reviews-1-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-2-reviews-1-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-2-reviews-1-rating"
                   id="id_products-2-reviews-1-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-2-reviews-1-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-2-reviews-1-DELETE"
                   id="id_products-2-reviews-1-DELETE">
                </td>
              </tr>
            </tbody>
            <tbody>
              <tr>
                <th><label for="id_products-2-reviews-2-title">Title:</label></th>
                <td>
                  <input type="text" name="products-2-reviews-2-title" maxlength="255"
                   id="id_products-2-reviews-2-title">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-2-reviews-2-rating">Rating:</label></th>
                <td>
                  <input type="number" name="products-2-reviews-2-rating"
                   id="id_products-2-reviews-2-rating">
                </td>
              </tr>
              <tr>
                <th><label for="id_products-2-reviews-2-DELETE">Delete:</label></th>
                <td>
                  <input type="checkbox" name="products-2-reviews-2-DELETE"
                   id="id_products-2-reviews-2-DELETE">
                </td>
              </tr>
            </tbody>
          </table>
          <input type="hidden" name="products-2-reviews-TOTAL_FORMS"
           value="3" id="id_products-2-reviews-TOTAL_FORMS">
          <input type="hidden" name="products-2-reviews-INITIAL_FORMS"
           value="0" id="id_products-2-reviews-INITIAL_FORMS">
          <input type="hidden" name="products-2-reviews-MIN_NUM_FORMS"
           id="id_products-2-reviews-MIN_NUM_FORMS">
          <input type="hidden" name="products-2-reviews-MAX_NUM_FORMS"
           id="id_products-2-reviews-MAX_NUM_FORMS">
        </td>
      </tr>
      <tr>
        <th><label for="id_products-2-DELETE">Delete:</label></th>
        <td>
          <input type="checkbox" name="products-2-DELETE" id="id_products-2-DELETE">
        </td>
      </tr>
    </tbody>
  </table>
  <input type="hidden" name="products-TOTAL_FORMS" value="3" id="id_products-TOTAL_FORMS">
  <input type="hidden" name="products-INITIAL_FORMS" value="1" id="id_products-INITIAL_FORMS">
  <input type="hidden" name="products-MIN_NUM_FORMS" id="id_products-MIN_NUM_FORMS">
  <input type="hidden" name="products-MAX_NUM_FORMS" id="id_products-MAX_NUM_FORMS">
</div>""",
        )

    def test_invalid_field_data(self):
        """A field's data (Review.title) is too long."""
        data = {
            "name": "Best Buy!",
            "products-0-name": "TV!",
            "products-0-reviews-0-title": "A" * 256,
            "products-0-reviews-0-rating": "9",
            "products-TOTAL_FORMS": 3,
            "products-INITIAL_FORMS": 1,
            "products-0-reviews-TOTAL_FORMS": 3,
            "products-0-reviews-INITIAL_FORMS": 1,
            "products-1-reviews-TOTAL_FORMS": 3,
            "products-1-reviews-INITIAL_FORMS": 0,
            "products-2-reviews-TOTAL_FORMS": 3,
            "products-2-reviews-INITIAL_FORMS": 0,
        }
        form = StoreForm(data)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["products"],
            ["Ensure this value has at most 255 characters (it has 256)."],
        )

    def test_all_missing_data(self):
        """
        An embedded model array field (reviews) with all data missing triggers
        a required error.
        """
        store = Store.objects.create(
            name="Best Buy",
            products=[Product(name="TV", reviews=[Review(title="Great", rating=9)])],
        )
        data = {
            "name": "Best Buy!",
            "products-0-name": "TV!",
            "products-0-reviews-0-title": "",
            "products-0-reviews-0-rating": "",
            "products-TOTAL_FORMS": 3,
            "products-INITIAL_FORMS": 1,
            "products-0-reviews-TOTAL_FORMS": 3,
            "products-0-reviews-INITIAL_FORMS": 1,
            "products-1-reviews-TOTAL_FORMS": 3,
            "products-1-reviews-INITIAL_FORMS": 0,
            "products-2-reviews-TOTAL_FORMS": 3,
            "products-2-reviews-INITIAL_FORMS": 0,
        }
        form = StoreForm(data, instance=store)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["products"], ["This field is required.", "This field is required."]
        )

    def test_rendering(self):
        form = StoreForm()
        self.assertHTMLEqual(
            str(form.fields["products"].get_bound_field(form, "products")),
            """
<table>
<tbody>
  <tr>
    <th><label for="id_products-0-name">Name:</label></th>
    <td>
      <input type="text" name="products-0-name" maxlength="255" id="id_products-0-name">
    </td>
  </tr>
  <tr>
    <th><label for="id_products-0-reviews">Reviews:</label></th>
    <td>
      <table>
        <tbody>
          <tr>
            <th><label for="id_products-0-reviews-0-title">Title:</label></th>
            <td>
              <input type="text" name="products-0-reviews-0-title" maxlength="255"
               id="id_products-0-reviews-0-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-0-reviews-0-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-0-reviews-0-rating"
               id="id_products-0-reviews-0-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-0-reviews-0-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-0-reviews-0-DELETE"
               id="id_products-0-reviews-0-DELETE">
            </td>
          </tr>
        </tbody>
        <tbody>
          <tr>
            <th><label for="id_products-0-reviews-1-title">Title:</label></th>
            <td>
              <input type="text" name="products-0-reviews-1-title" maxlength="255"
               id="id_products-0-reviews-1-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-0-reviews-1-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-0-reviews-1-rating"
               id="id_products-0-reviews-1-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-0-reviews-1-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-0-reviews-1-DELETE"
               id="id_products-0-reviews-1-DELETE">
            </td>
          </tr>
        </tbody>
        <tbody>
          <tr>
            <th><label for="id_products-0-reviews-2-title">Title:</label></th>
            <td>
              <input type="text" name="products-0-reviews-2-title" maxlength="255"
               id="id_products-0-reviews-2-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-0-reviews-2-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-0-reviews-2-rating"
               id="id_products-0-reviews-2-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-0-reviews-2-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-0-reviews-2-DELETE"
               id="id_products-0-reviews-2-DELETE">
            </td>
          </tr>
        </tbody>
      </table>
      <input type="hidden" name="products-0-reviews-TOTAL_FORMS" value="3"
       id="id_products-0-reviews-TOTAL_FORMS"><input type="hidden"
       name="products-0-reviews-INITIAL_FORMS" value="0"
       id="id_products-0-reviews-INITIAL_FORMS">
     <input type="hidden" name="products-0-reviews-MIN_NUM_FORMS" value="0"
       id="id_products-0-reviews-MIN_NUM_FORMS">
       <input type="hidden" name="products-0-reviews-MAX_NUM_FORMS"
         value="1000" id="id_products-0-reviews-MAX_NUM_FORMS">
    </td>
  </tr>
  <tr>
    <th><label for="id_products-0-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="products-0-DELETE" id="id_products-0-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_products-1-name">Name:</label></th>
    <td>
      <input type="text" name="products-1-name" maxlength="255" id="id_products-1-name">
    </td>
  </tr>
  <tr>
    <th><label for="id_products-1-reviews">Reviews:</label></th>
    <td>
      <table>
        <tbody>
          <tr>
            <th><label for="id_products-1-reviews-0-title">Title:</label></th>
            <td>
              <input type="text" name="products-1-reviews-0-title" maxlength="255"
               id="id_products-1-reviews-0-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-1-reviews-0-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-1-reviews-0-rating"
               id="id_products-1-reviews-0-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-1-reviews-0-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-1-reviews-0-DELETE"
               id="id_products-1-reviews-0-DELETE">
            </td>
          </tr>
        </tbody>
        <tbody>
          <tr>
            <th><label for="id_products-1-reviews-1-title">Title:</label></th>
            <td>
              <input type="text" name="products-1-reviews-1-title" maxlength="255"
               id="id_products-1-reviews-1-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-1-reviews-1-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-1-reviews-1-rating"
               id="id_products-1-reviews-1-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-1-reviews-1-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-1-reviews-1-DELETE"
               id="id_products-1-reviews-1-DELETE">
            </td>
          </tr>
        </tbody>
        <tbody>
          <tr>
            <th><label for="id_products-1-reviews-2-title">Title:</label></th>
            <td>
              <input type="text" name="products-1-reviews-2-title" maxlength="255"
               id="id_products-1-reviews-2-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-1-reviews-2-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-1-reviews-2-rating"
               id="id_products-1-reviews-2-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-1-reviews-2-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-1-reviews-2-DELETE"
               id="id_products-1-reviews-2-DELETE">
            </td>
          </tr>
        </tbody>
      </table>
      <input type="hidden" name="products-1-reviews-TOTAL_FORMS" value="3"
       id="id_products-1-reviews-TOTAL_FORMS"><input type="hidden"
       name="products-1-reviews-INITIAL_FORMS" value="0"
       id="id_products-1-reviews-INITIAL_FORMS">
     <input type="hidden" name="products-1-reviews-MIN_NUM_FORMS" value="0"
       id="id_products-1-reviews-MIN_NUM_FORMS">
       <input type="hidden" name="products-1-reviews-MAX_NUM_FORMS"
         value="1000" id="id_products-1-reviews-MAX_NUM_FORMS">
    </td>
  </tr>
  <tr>
    <th><label for="id_products-1-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="products-1-DELETE" id="id_products-1-DELETE">
    </td>
  </tr>
</tbody>
<tbody>
  <tr>
    <th><label for="id_products-2-name">Name:</label></th>
    <td>
      <input type="text" name="products-2-name" maxlength="255" id="id_products-2-name">
    </td>
  </tr>
  <tr>
    <th><label for="id_products-2-reviews">Reviews:</label></th>
    <td>
      <table>
        <tbody>
          <tr>
            <th><label for="id_products-2-reviews-0-title">Title:</label></th>
            <td>
              <input type="text" name="products-2-reviews-0-title" maxlength="255"
               id="id_products-2-reviews-0-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-2-reviews-0-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-2-reviews-0-rating"
               id="id_products-2-reviews-0-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-2-reviews-0-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-2-reviews-0-DELETE"
               id="id_products-2-reviews-0-DELETE">
            </td>
          </tr>
        </tbody>
        <tbody>
          <tr>
            <th><label for="id_products-2-reviews-1-title">Title:</label></th>
            <td>
              <input type="text" name="products-2-reviews-1-title" maxlength="255"
               id="id_products-2-reviews-1-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-2-reviews-1-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-2-reviews-1-rating"
               id="id_products-2-reviews-1-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-2-reviews-1-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-2-reviews-1-DELETE"
               id="id_products-2-reviews-1-DELETE">
            </td>
          </tr>
        </tbody>
        <tbody>
          <tr>
            <th><label for="id_products-2-reviews-2-title">Title:</label></th>
            <td>
              <input type="text" name="products-2-reviews-2-title" maxlength="255"
               id="id_products-2-reviews-2-title">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-2-reviews-2-rating">Rating:</label></th>
            <td>
              <input type="number" name="products-2-reviews-2-rating"
               id="id_products-2-reviews-2-rating">
            </td>
          </tr>
          <tr>
            <th><label for="id_products-2-reviews-2-DELETE">Delete:</label></th>
            <td>
              <input type="checkbox" name="products-2-reviews-2-DELETE"
               id="id_products-2-reviews-2-DELETE">
            </td>
          </tr>
        </tbody>
      </table>
      <input type="hidden" name="products-2-reviews-TOTAL_FORMS" value="3"
       id="id_products-2-reviews-TOTAL_FORMS"><input type="hidden"
       name="products-2-reviews-INITIAL_FORMS" value="0"
       id="id_products-2-reviews-INITIAL_FORMS">
     <input type="hidden" name="products-2-reviews-MIN_NUM_FORMS" value="0"
       id="id_products-2-reviews-MIN_NUM_FORMS">
       <input type="hidden" name="products-2-reviews-MAX_NUM_FORMS"
         value="1000" id="id_products-2-reviews-MAX_NUM_FORMS">
    </td>
  </tr>
  <tr>
    <th><label for="id_products-2-DELETE">Delete:</label></th>
    <td>
      <input type="checkbox" name="products-2-DELETE" id="id_products-2-DELETE">
    </td>
  </tr>
</tbody>
</table>
<input type="hidden" name="products-TOTAL_FORMS" value="3"
id="id_products-TOTAL_FORMS"><input type="hidden" name="products-INITIAL_FORMS"
value="0" id="id_products-INITIAL_FORMS">
<input type="hidden" name="products-MIN_NUM_FORMS" value="0" id="id_products-MIN_NUM_FORMS">
<input type="hidden" name="products-MAX_NUM_FORMS" value="1000" id="id_products-MAX_NUM_FORMS">""",
        )



from bson import json_util
from django.apps import apps
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, connections, router

from django_mongodb_backend.utils import model_has_encrypted_fields


class Command(BaseCommand):
    help = "Shows the encrypted_fields_map for use in AutoEncryptionOpts."

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help='Specifies the database to use. Defaults to the "default" database.',
        )

    def handle(self, *args, **options):
        db = options["database"]
        connection = connections[db]
        connection.ensure_connection()
        encrypted_fields_map = {}
        with connection.schema_editor() as editor:
            for app_config in apps.get_app_configs():
                for model in router.get_migratable_models(app_config, db):
                    if model_has_encrypted_fields(model) and model._meta.can_migrate(connection):
                        fields = editor._get_encrypted_fields(model, create_data_keys=False)
                        encrypted_fields_map[model._meta.db_table] = fields
        self.stdout.write(json_util.dumps(encrypted_fields_map, indent=4))
